from __future__ import division,print_function

import numpy as np
from math import pi, atan
# If scipy is available, use its optimization function, otherwise, 
# use our implementation (for packaging purposes)
try:
    import scipy.optimize as optimize
except ImportError:
    import PDSim.misc.scipylike as optimize
    
import matplotlib.pyplot as plt

N = 61
#e_mat=[0.2,0.25,0.3,0.35,0.4,0.5,0.6,0.7,0.8,0.9];

phi_star = pi

def TwoDGriddedIntegrate(I,N):
    # Average the center of each cell based on its neighboring nodes
    return np.sum(np.sum((I[0:N-1,0:N-1]+I[1:N,0:N-1]+I[0:N-1,1:N]+I[1:N,1:N])))/4
    
def TwoDGriddedIntegrate2(PHI,Y,I):
    
    #Integrate along phi direction for each y, then do a trapezoidal integration of each of the y
    plt.plot(Y[1,:],np.trapz(I,PHI,axis = 0))
    plt.show()
    return np.trapz(np.trapz(I,PHI,axis = 0),Y[1,:])

def OBJECTIVE(phi_star, epsilon, plot = False, output = False):

    PHI = np.tile(np.linspace(0,phi_star,N).T,(N,1)).T
    Y = np.tile(np.linspace(0,1,N),(N,1))
    
    dPHI = phi_star/(N-1)
    dY = 1/(N-1)
    sinPHI=np.sin(PHI)
    P = 0*PHI
    Pnew = 0*PHI
    f = 0*PHI
    df = 0*PHI
    _lambda = 1
    change = 999    
    
    eps=1e-6;
    count=0;
    while (change>eps):
        
        #Calculate geometric parameters
        H=1+epsilon*np.cos(PHI);
        H3=H**3;
        
        #Coefficients
        A=H3[2:N,1:N-1]
        B=H3[0:N-2,1:N-1]
        C=H3[1:N-1,1:N-1]

        #Calculate residuals
        f[1:N-1,1:N-1] = -(4*A+4*B+2*_lambda*dPHI**2/dY**2*C)*P[1:N-1,1:N-1]+(3*A+B)*P[2:N,1:N-1]+(A+3*B)*P[0:N-2,1:N-1]+(_lambda**2*dPHI**2/dY**2*C)*(P[1:N-1,2:N]+P[1:N-1,0:N-2])+24*dPHI**2*epsilon*sinPHI[1:N-1,1:N-1]

        #Calculate derivative
        df[1:N-1,1:N-1]=-(4*A+4*B+2*_lambda*dPHI**2/dY**2*C);

        #Evaluate P_new=P_old-f/dfdP
        P[1:N-1,1:N-1]=P[1:N-1,1:N-1]-f[1:N-1,1:N-1]/df[1:N-1,1:N-1];

        #Evaluate change
        change=np.max(np.max(np.abs(f[1:N-1,1:N-1]/df[1:N-1,1:N-1])));

        if count % 1000 == 0:
            print(change)
            
        count += 1
        
    if output:
        Wx=dY*dPHI*np.sum(np.sum(np.sin(PHI)*P))
        Wz=-dY*dPHI*np.sum(np.sum(np.cos(PHI)*P))
        Wr = np.sqrt(Wx**2+Wz**2)
        PHI_angle = atan(Wx/Wz)
        B_j = 1/(pi*Wr)
        DPDPHI = 0*Y
        DPDPHI[0:N-2,0:N] = (P[1:N-1,0:N]-P[0:N-2,0:N])/(dPHI)
        DPDPHI[N-1:N-1,0:N] = (P[N-1:N,0:N]-P[N-2:N-2,0:N])/(dPHI)
        
        integrand = 1/H
        #integrand = H/2*DPDPHI+1/H
        
        Fb1 = dPHI*dY*np.sum(np.sum(integrand))
        Fb2 = dPHI*dY*TwoDGriddedIntegrate(integrand,N)
        Fb3 = TwoDGriddedIntegrate2(PHI,Y,integrand)
        mu_rb_c = Fb3/Wr # mu*r_b/c
        print('Fb1,Fb2,Fb3',Fb1,Fb2,Fb3)
        print('B_j', B_j)
        print('mu*rb/c', mu_rb_c)
        #print 'mu*rb/c', mu_rb_c/12.8
        print('PHI_angle', PHI_angle/pi*180)
        plt.contour(PHI,Y,H/2*DPDPHI+1/H)
        plt.show()
    
    if plot:
        plt.contour(PHI,Y,P,30)
        plt.show()
    
    return np.sum(3*P[N-1,N//2+1]-4*P[N-2,N//2+1]+P[N-3,N//2+1])/(2*dPHI)
        
if __name__=='__main__':
    print(optimize.newton.__doc__); quit()
    phi_star = optimize.newton(OBJECTIVE, pi, args = (0.6,), tol = 0.004)
    
    OBJECTIVE(phi_star,0.6,plot = True, output = True)



from __future__ import print_function

from math import log10, isnan

def error_ascii_bar(val, tol, N = 50, min_p = -5, max_p = 2):

    assert log10(tol) > min_p
    
    base_string = list('||'+'.'*N+'||')
    
    if isnan(val):
        return ' XXX VAL is NAN XXX '

    # Add the bar for the tolerance
    i = int(N*(log10(tol)-min_p)/(max_p-min_p)+1)
    base_string[i] = '|'
    
    # Add the value
    i = int(N*(log10(val)-min_p)/(max_p-min_p)+1)
    # Constrain the value so that it is definitely in range
    i = max(1, i)
    i = min(i, len(base_string)-2)

    base_string[i] = '@'
    
    base_string = ''.join(base_string)
    
    return base_string

if __name__=='__main__':
    print(error_ascii_bar(1,1e-3))
    print(error_ascii_bar(0.5,1e-3))
    print(error_ascii_bar(0.1,1e-3))
    print(error_ascii_bar(0.0011,1e-3))
    print(error_ascii_bar(1e-10,1e-3))
    print(error_ascii_bar(1e10,1e-3))



from __future__ import division, absolute_import, print_function

import math
from math import pi
from time import clock
import inspect
import six

##--  Package imports  --
from PDSim.flow import flow,flow_models
from .containers import STATE_VARS_TM, CVArrays, ControlVolumeCollection,TubeCollection
from PDSim.flow.flow import FlowPathCollection
from . import integrators
from PDSim.misc.datatypes import arraym, empty_arraym
import PDSim.core.callbacks
from PDSim.misc.error_bar import error_ascii_bar

##-- Non-package imports  --
import numpy as np

# If scipy is available, use its optimization function, otherwise, 
# use our implementation (for packaging purposes)
try:
    from scipy.integrate import trapz
except ImportError:
    from PDSim.misc.scipylike import trapz

import h5py

## matplotlib imports
import matplotlib.pyplot as plt
import pylab

## CoolProp imports
from CoolProp.State import State

#def _pickle_method(method):
    #func_name = method.im_func.__name__
    #obj = method.im_self
    #cls = method.im_class
    #return _unpickle_method, (func_name, obj, cls)

#def _unpickle_method(func_name, obj, cls):
    #for cls in cls.mro():
        #try:
            #func = cls.__dict__[func_name]
        #except KeyError:
            #pass
        #else:
            #break
        #return func.__get__(obj, cls)

#import copy_reg
#import types
#copy_reg.pickle(types.MethodType, _pickle_method, _unpickle_method)

# An empty class for storage
class struct(object):
    pass    
    
class IntegratorMixin(object):
    """
    This class contains the methods that will be merged with one of the system of ODE integrators
    and includes the methods that are specific to PDSim
    """
    def __init__(self, sim, x_state):
        self.sim = sim
        self.x_state = x_state
    
    def get_initial_array(self):
        # Get the beginning of the cycle configured
        # Put a copy of the values into the matrices
        xold = self.x_state.copy()
        # Add zeros for the valves as they are assumed to start closed and at rest
        if self.sim.__hasValves__:
            xold.extend(empty_arraym(len(self.sim.Valves)*2))
        self.sim._put_to_matrices(xold, 0)
        return xold
        
    def premature_termination(self):
        # Once every 100 steps check if you are supposed to abort
        if self.sim._check_cycle_abort(self.Itheta):
            return 'abort'
        else:
            return False
        
    def pre_step_callback(self):
        # Call the step callback if provided
        if self.sim.callbacks.step_callback is not None:
            self.h = self.sim.callbacks.step_callback(self.t0, self.h, self.Itheta)
            disable = self.sim.callbacks.step_callback.disable_adaptive
            
            # If we don't want to actually do the step, rather just copy the values
            # (for instance at the merging angle for scroll machines), we set
            # stepAccepted to True and move on
            if disable == 'no_integrate':
                self.stepAccepted = True
                # Retrieve the array of values based on the values set in step_callback
                x = self.sim._get_from_matrices(self.Itheta)
                # Updates the state, calculates the volumes, prepares all the things needed for derivatives
                # The crank angle must be evaluated after theta_d in order to ensure that the new volumes are used
                # The newly calculated values are used
                self.sim.core.properties_and_volumes(self.sim.CVs.exists_CV, self.t0+self.h+1e-10, STATE_VARS_TM, x)
                self.xold = x.copy()
                self.xnew = x.copy()
                self.__store_values()
                
            x = self.sim._get_from_matrices(self.Itheta)
                
            if disable != False and x.all_finite():
                self.xold = self.sim._get_from_matrices(self.Itheta)
                
            # The integrator only cares whether disable is true or not, convert to true or false
            if disable != False:
                self.disableAdaptive = True
            else:
                self.disableAdaptive = False
    
    def __store_values(self):
        """ Private method that stores the values in the internal data structure """
        self.sim.t[self.Itheta] = self.t0
        self.sim._put_to_matrices(self.xold, self.Itheta)
        flows = self.sim.Flows.get_deepcopy()
        # If the index we want to fill is beyond the length of FlowStorage, we append it,
        # otherwise, we replace that entry in the list
        if self.Itheta > len(self.sim.FlowStorage)-1:
            self.sim.FlowStorage.append(flows)
        else:
            self.sim.FlowStorage[self.Itheta] = flows
        
    def post_deriv_callback(self):
        self.__store_values()
        
    def post_step_callback(self): pass
    
    def derivs(self, t, x):
        return self.sim.derivs(t, x)
        
    def post_integration(self):
        """
        Run this at the end
        """
        # Cache the values
        self.sim.derivs(self.t0, self.xold)
        self.__store_values()
        
        V, dV = self.sim.CVs.volumes(self.t0)
        Nexist = self.sim.CVs.Nexist
        if sorted(self.sim.stateVariables) == ['D','T']:
            self.sim.CVs.updateStates('T',self.xnew[0:Nexist],'D',self.xnew[Nexist:2*Nexist])
        elif sorted(self.sim.stateVariables) == ['M','T']:
            self.sim.CVs.updateStates('T',self.xnew[0:Nexist],'D',self.xnew[Nexist:2*Nexist]/V)
        else:
            raise NotImplementedError
            
class EulerIntegrator(IntegratorMixin, integrators.AbstractSimpleEulerODEIntegrator):
    """ 
    Mixin class using the functions defined in IntegratorMixin and the generalized simple ODE 
    """
    def __init__(self, sim, x_state):
        IntegratorMixin.__init__(self, sim, x_state)
        
class HeunIntegrator(IntegratorMixin, integrators.AbstractHeunODEIntegrator):
    """ 
    Mixin class using the functions defined in IntegratorMixin and the generalized Heun ODE integrator
    """
    def __init__(self, sim, x_state):
        IntegratorMixin.__init__(self, sim, x_state)
        
class RK45Integrator(IntegratorMixin, integrators.AbstractRK45ODEIntegrator):
    """ 
    Mixin class using the functions defined in IntegratorMixin and the generalized RK45 integrator
    """
    def __init__(self, sim, x_state):
        IntegratorMixin.__init__(self, sim, x_state)

class PDSimCore(object):
    """
    This is the main driver class for the model
    
    This class is not intended to be run on its own.  It must be subclassed and extended to provide functions for mass flow, etc. 
    
    The basic order of steps that should be followed can be summarized as
    
    #. Instantiate the subclass of PDSimCore
    #. Add each of the control volumes
    #. Add each of the tubes
    #. Add all the flow models between CV and tubes
    #. Add valves (if applicable)
    #. Connect the callbacks for heat transfer, step, etc.
    #. Run the model

    """
    def __init__(self,stateVariables=None):
        """
        Initialization of the PDSimCore
        
        Parameters
        ----------
        stateVariables : mutable object [list or tuple], optional
            list of keys for the state variables to be used.  Current options are 'T','D' or 'T','M'.  Default state variables are 'T','M'
        """
        #Initialize the containers to be empty
        
        #: The Valves container class
        self.Valves=[]
        #: The :class:`ControlVolumeCollection <PDSim.core.containers.ControlVolumeCollection>` instance
        #: that contains all the control volumes in the machine
        self.CVs=ControlVolumeCollection()
        #: The :class:`FlowPathCollection <PDSim.flow.flow.FlowPathCollection>` 
        #: instance
        self.Flows=FlowPathCollection()
        #: A :class:`list` that contains copies of the 
        #: :class:`FlowPathCollection <PDSim.flow.flow.FlowPathCollection>` 
        #: at each crank angle
        self.FlowStorage=[]
        
        self.Tubes = TubeCollection()
        self.Tlumps = np.zeros((1,1))
        self.steps = []
        self.__hasValves__ = False
        
        #  A storage of the initial state vector
        self.xstate_init = None
        
        # A storage of the initial valves vector
        if isinstance(stateVariables,(list,tuple)):
            self.stateVariables=list(stateVariables)
        else:
            self.stateVariables=['T','M']
        self._want_abort = False
        
        #  Build a structure to hold all the callbacks
        self.callbacks = PDSim.core.callbacks.CallbackContainer()
        
        #  Build a dummy class to hold information from the solvers
        class dummy: pass
        self.solvers = dummy()
        self.solvers.lump_eb_history = []
        self.solvers.hdisc_history = []
        self.solvers.initial_states_history = []
        
        self.verbosity = 0
        
        self.summary = dummy()
    
    def _check(self):
        """
        Do some checking before we start the run.
        
        Here we check:
        
        * Inlet state viscosity and conductivity must be greater than zero
        """
        
        if self.inlet_state.get_visc() < 0:
            raise ValueError('Your inlet state viscosity is less than zero. Invalid fluid: ' +self.inlet_state.Fluid)
        if self.inlet_state.get_cond() < 0:
            raise ValueError('Your inlet state conductivity is less than zero. Invalid fluid: '+self.inlet_state.Fluid)
            
    def _get_from_matrices(self,i):
        """
        Get values back from the matrices and reconstruct the state variable list
        """
        if self.__hasLiquid__==True:
            raise NotImplementedError
        else:
            VarList=np.array([])
            exists_indices = np.array(self.CVs.exists_indices)
            for s in self.stateVariables:
                if s=='T':
                    VarList = np.append(VarList, self.T[exists_indices,i])
                elif s=='D':
                    VarList = np.append(VarList, self.rho[exists_indices,i])
                elif s=='M':
                    VarList = np.append(VarList, self.m[exists_indices,i])
                else:
                    raise KeyError
            
            if self.__hasValves__:

                ValveList = list(self.xValves[:,i])
                VarList = np.append( VarList,ValveList)

            return arraym(VarList)
        
    def _statevars_to_dict(self,x):
        d={}
        for iS,s in enumerate(self.stateVariables):
            x_=(x[iS*self.CVs.Nexist:self.CVs.Nexist*(iS+1)])
            if s=='T':
                d['T']=x_
            elif s=='D':
                d['D']=x_
            elif s=='M':
                d['M']=x_
        return d
        
    def _put_to_matrices(self,x,i):
        """
        Take a state variable list and put back in numpy matrices
        """
        exists_indices=self.CVs.exists_indices
        Nexist = self.CVs.Nexist
        Ns = len(self.stateVariables)
        if self.__hasLiquid__==True:
            raise NotImplementedError
#             self.T[:,i]=x[0:self.NCV]
#             self.m[:,i]=x[self.NCV:2*self.NCV]
        else: # self.__hasLiquid__==False
            for iS, s in enumerate(self.stateVariables):
                if s=='T':
                    self.T[exists_indices, i] = x[iS*self.CVs.Nexist:self.CVs.Nexist*(iS+1)]
                elif s=='D':
                    self.rho[exists_indices, i] = x[iS*self.CVs.Nexist:self.CVs.Nexist*(iS+1)]
                elif s=='M':
                    self.m[exists_indices, i] = x[iS*self.CVs.Nexist:self.CVs.Nexist*(iS+1)]
            #Left over terms are for the valves
            if self.__hasValves__:
                self.xValves[list(range(len(self.Valves)*2)), i] = arraym(x[Ns*Nexist:len(x)])
            
            # In the first iteration, self.core has not been filled, so do not 
            # overwrite with the values in self.core.m and self.core.rho
            if self.core.m[0] > 0.0 :
                self.m[exists_indices, i] = self.core.m
            if self.core.rho[0] > 0.0 :
                self.rho[exists_indices, i] = self.core.rho
            
            self.V[exists_indices, i] = self.core.V
            self.dV[exists_indices, i] = self.core.dV
            self.p[exists_indices, i] = self.core.p
            self.h[exists_indices, i] = self.core.h
            self.Q[exists_indices,i] = self.core.Q
    
    def _postprocess_flows(self):
        """
        In this private method, the flows from each of the flow nodes are summed for 
        each step of the revolution, and then averaged flow rates are calculated.
        """
        
        def sum_flows(key,Flows):
            """
            Sum all the terms for a given flow key.  
            
            Flows "into" the node are positive, flows out of the 
            node are negative
            
            Use the code in the Cython module
            """
            return flow.sumterms_given_CV(key, Flows)
            
        def collect_keys(Tubes,Flows):
            """
            Get all the keys for a given collection of flow elements
            """
            keys=[]
            for Tube in Tubes:
                if Tube.key1 not in keys:
                    keys.append(Tube.key1)
                if Tube.key2 not in keys:
                    keys.append(Tube.key2)
            for Flow in Flows:
                if Flow.key1 not in keys:
                    keys.append(Flow.key1)
                if Flow.key2 not in keys:
                    keys.append(Flow.key2)
            return keys
        
        #  Get all the nodes that can exist for tubes and CVs
        keys=collect_keys(self.Tubes,self.Flows)
        
        #  Get the instantaneous net flow through each node
        #  and the averaged mass flow rate through each node
        self.FlowsProcessed=struct()
        self.FlowsProcessed.summed_mdot={}
        self.FlowsProcessed.summed_mdoth={}
        self.FlowsProcessed.mean_mdot={}
        self.FlowsProcessed.integrated_mdoth={}
        self.FlowsProcessed.integrated_mdot={}
        self.FlowsProcessed.t=self.t[0:self.Ntheta]

        for key in keys:
            # Empty container numpy arrays
            self.FlowsProcessed.summed_mdot[key]=np.zeros((self.Ntheta,))
            self.FlowsProcessed.summed_mdoth[key]=np.zeros((self.Ntheta,))
            
            assert self.Ntheta == len(self.FlowStorage)
            for i in range(self.Ntheta):
                mdot,mdoth=sum_flows(key,self.FlowStorage[i])
                self.FlowsProcessed.summed_mdot[key][i]=mdot
                self.FlowsProcessed.summed_mdoth[key][i]=mdoth
            
            # All the calculations here should be done in the time domain,
            # rather than crank angle.  So convert angle to time by dividing 
            # by omega, the rotational speed in rad/s.
            trange = self.t[self.Ntheta-1]-self.t[0]
            # integrated_mdoth has units of kJ/rev * f [Hz] --> kJ/s or kW
            self.FlowsProcessed.integrated_mdoth[key]=trapz(self.FlowsProcessed.summed_mdoth[key], 
                                                            self.t[0:self.Ntheta]/self.omega)*self.omega/trange
            # integrated_mdot has units of kg/rev * f [Hz] --> kg/s
            self.FlowsProcessed.integrated_mdot[key]=trapz(self.FlowsProcessed.summed_mdot[key], 
                                                           self.t[0:self.Ntheta]/self.omega)*self.omega/trange
            self.FlowsProcessed.mean_mdot[key]=np.mean(self.FlowsProcessed.integrated_mdot[key])
            
        # Special-case the tubes.  Only one of the nodes can have flow.  
        # The other one is invariant because it is quasi-steady.
        for Tube in self.Tubes:
            mdot1 = self.FlowsProcessed.mean_mdot[Tube.key1]
            mdot2 = self.FlowsProcessed.mean_mdot[Tube.key2]
            mdot_i1 = self.FlowsProcessed.integrated_mdot[Tube.key1]
            mdot_i2 = self.FlowsProcessed.integrated_mdot[Tube.key2]
            mdoth_i1 = self.FlowsProcessed.integrated_mdoth[Tube.key1]
            mdoth_i2 = self.FlowsProcessed.integrated_mdoth[Tube.key2]
            #Swap the sign so the sum of the mass flow rates is zero
            self.FlowsProcessed.mean_mdot[Tube.key1] -= mdot2
            self.FlowsProcessed.mean_mdot[Tube.key2] -= mdot1
            self.FlowsProcessed.integrated_mdot[Tube.key1] -= mdot_i2
            self.FlowsProcessed.integrated_mdot[Tube.key2] -= mdot_i1
            self.FlowsProcessed.integrated_mdoth[Tube.key1] -= mdoth_i2
            self.FlowsProcessed.integrated_mdoth[Tube.key2] -= mdoth_i1
            
            #For each tube, update the flow going through it
            #Tube.mdot is always a positive value
            Tube.mdot = max(abs(mdot1), abs(mdot2))
            
        self.mdot = self.FlowsProcessed.mean_mdot[self.key_inlet]
        
        self.FlowsProcessed.collected_data = []
                
        for i, Flow in enumerate(self.Flows):
             
            mdot = np.array([Flows[i].mdot for Flows in self.FlowStorage])
            edot = np.array([Flows[i].edot for Flows in self.FlowStorage])
            
            data = dict(key1 = Flow.key1,
                        key2 = Flow.key2,
                        fcn = Flow.MdotFcn_str,
                        mdot = mdot,
                        edot = edot,
                        mdot_average = np.trapz(mdot, self.t[0:self.Ntheta])/(self.t[self.Ntheta-1]-self.t[0]),
                        Edot_average = np.trapz(edot, self.t[0:self.Ntheta])/(self.t[self.Ntheta-1]-self.t[0]) 
                        )
                
            self.FlowsProcessed.collected_data.append(data)
            
    def _postprocess_HT(self):
        """
        Postprocess the heat transfer terms
        
        Here we
        calculate the mean heat transfer rate over the course of the cycle 
        """
        self.HTProcessed=struct()
        r = list(range(self.Ntheta))
        
        #Remove all the NAN placeholders and replace them with zero values
        self.Q[np.isnan(self.Q)] = 0.0
        
        #Sum at each step of the revolution
        self.HTProcessed.summed_Q = np.sum(self.Q, axis = 0) #kW
        
        #Get the mean heat transfer rate
        self.HTProcessed.mean_Q = trapz(self.HTProcessed.summed_Q[r], self.t[r])/(self.t[self.Ntheta-1]-self.t[0])
    
    
    def guess_outlet_temp(self, inlet_state, p_outlet, eta_a=0.7):
        """ 
        Function to guess outlet temperature
        
        Using a guess value for the adiabatic efficiency, calculate the guessed
        outlet temperature.  In compressor mode, the adiabatic efficiency is defined by
        
        .. math::
        
            \eta_a = \\frac{h_{2s}-h_1}{h_2-h_1}
            
        and in expander mode it is defined by
        
        .. math::
        
            \eta_a = \\frac{h_2-h_1}{h_{2s}-h_1}
            
        This function can also be overloaded by the subclass in order to 
        implement a different guess method
        """
        
        h1 = inlet_state.h
        out_state = inlet_state.copy()
        out_state.update(dict(S = inlet_state.s, P = p_outlet))
        h2s = out_state.h
        if p_outlet > inlet_state.p:
            # Compressor Mode
            h2 = h1 + (h2s-h1)/eta_a
        else:
            # Expander Mode
            h2 = h1 + (h2s-h1)*eta_a
        out_state.update(dict(H = h2, P = p_outlet))
        return out_state.T
    
    def reset_initial_state(self):
        """
        Reset the initial state of the core class, typically after doing a 
        preconditioning run
        """
        
        for k,CV in zip(self.CVs.keys,self.CVs.CVs):
            if k in self.exists_CV_init:
                CV.exists = True
            else:
                CV.exists = False

        #Update the existence of each of the CV
        self.update_existence()
             
        #Only the State variables, not the valves
        self.x_state = self.xstate_init
        #Make a copy
        x = self.xstate_init.copy()
        #Add the values from the valves
        if self.__hasValves__:
            x.extend(empty_arraym(2*len(self.Valves)))
        self._put_to_matrices(x, 0)
        #Reset the temporary variables
        self.xstate_init = None
        self.exists_CV_init = None
                
    def update_existence(self):
        """
        Update existence flags for Tubes and control volumes
        
        This function is required to be called when the existence of any control
        volume or tube changes in order to ensure that internal flags are set
        properly
        """
        
        # Update the existence flags in all the control volumes
        self.CVs.rebuild_exists()
        
        # Update the array of enthalpies in the tubes
        self.Tubes.update_existence(self.CVs.Nexist)
        
        # Update the existence of each of the flows
        self.Flows.update_existence(self)
        
        # Update the sizes of the internal arrays in self.core  
        self.core.update_size(self.CVs.Nexist)
        
    def add_flow(self,FlowPath):
        """
        Add a flow path to the model
        
        Parameters
        ----------
        FlowPath : :class:`FlowPath <PDSim.flow.flow.FlowPath>`  instance
            An initialized flow path 
        """
        #Add FlowPath instance to the list of flow paths
        self.Flows.append(FlowPath)
        
    def add_CV(self,CV):
        """
        Add a control volume to the model
        
        Parameters
        ----------
        CV : :class:`ControlVolume <PDSim.core.containers.ControlVolume>` instance
            An initialized control volume
        """
        
        if CV.key in self.CVs.keys:
            raise KeyError('Sorry but the key for your Control Volume ['+CV.key+'] is already in use')
        
        #Add the CV to the collection
        self.CVs.add(CV)
        self.CVs.rebuild_exists()
        
    def add_tube(self,Tube):
        """
        Add a tube to the model.
        
        Parameters
        ----------
        Tube : :class:`Tube <PDSim.core.containers.Tube>` instance
            An initialized tube.
        """
        #Add it to the list
        self.Tubes.append(Tube)
        self.Tubes.update()
        
    def add_valve(self,Valve):
        """
        Add a valve to the model.
        
        Parameters
        ----------
        Valve : :class:`ValveModel <PDSim.flow.flow_models.ValveModel>` instance
            An initialized valve.
        """
        #Add it to the list
        self.Valves.append(Valve)
        self.__hasValves__=True
        
    def pre_run(self, N = 40000):
        """
        This function gets called before the run begins.  It builds large matrices
        to store values, and does other initialization. 
        """
        # Build the full numpy arrays for temperature, volume, etc.
        self.t=np.zeros((N,))
        self.T=np.zeros((self.CVs.N,N))
        self.T.fill(np.nan)
        self.p=self.T.copy()
        self.h = self.T.copy()
        self.m = self.T.copy()
        self.V = self.T.copy()
        self.dV = self.T.copy()
        self.rho = self.T.copy()
        self.Q = self.T.copy()
        self.xValves = np.zeros((2*len(self.Valves),N))
        
        # Initialize the core class that contains the arrays and the derivs
        self.core = CVArrays(0)
        
        # Update the existence of all the control volumes
        self.update_existence()
        
        # Set a flag about liquid flooding
        self.__hasLiquid__ = False
        
    def pre_cycle(self, x0 = None):
        """
        This runs before the cycle is run but after pre_run has been called
        
        Parameters
        ----------
        x0 : :class:`arraym <PDSim.misc.datatypes.arraym>` instance
        """
        self.t.fill(np.nan)
        self.T.fill(np.nan)
        self.p.fill(np.nan)
        self.m.fill(np.nan)
        self.V.fill(np.nan)
        self.dV.fill(np.nan)
        self.rho.fill(np.nan)
        self.Q.fill(np.nan)
        
        self.FlowStorage=[]
        
        #Get the volumes at theta=0
        #Note: needs to occur in this function because V needed to calculate mass a few lines below
        VdV=[CV.V_dV(0.0,**CV.V_dV_kwargs) for CV in self.CVs.exists_CV]
        V,dV = zip(*VdV)
        
        self.t[0]=0
        
        # If x0 is provided, use its values to initialize the chamber states
        if x0 is None:
            # self.CVs.exists_indices is a list of indices of the CV with the same order of entries
            # as the entries in self.CVs.T
            self.T[self.CVs.exists_indices, 0] = self.CVs.T
            self.p[self.CVs.exists_indices, 0] = self.CVs.p
            self.rho[self.CVs.exists_indices, 0] = self.CVs.rho
            self.m[self.CVs.exists_indices, 0] = self.CVs.rho*arraym(V)
        else:
            #x0 is provided, but need to pad it out to include valve values
            x0_ = x0.copy()
            
            # If x0 is longer than the product of the number of state variables 
            # and CV in existence, the valve data is already included and must not be 
            # added to the array of independent variables
            if self.__hasValves__ and len(x0) == self.CVs.Nexist*len(self.stateVariables):
                #Load up the rest of the array with zeros since the valves start closed and at rest
                x0_.extend(empty_arraym(len(self.Valves)*2))
            self._put_to_matrices(x0_, 0)
        
        # Assume all the valves to be fully closed and stationary at the beginning of cycle
        self.xValves[:,0]=0
        
        self.Tubes_hdict={}
        for Tube in self.Tubes:
            self.Tubes_hdict[Tube.key1]=Tube.State1.get_h()
            self.Tubes_hdict[Tube.key2]=Tube.State2.get_h()
        
    def calc_boundary_work(self):
        """
        This method calculates the boundary work rate using a trapezoidal 
        integration of
        
        .. math::
        
            \\dot W_{pv} = -\int p\\frac{dV}{d\\theta}\\frac{\\omega}{2\\pi} d\\theta
            
        for all the control volumes and sets the parameter ``self.Wdot_pv`` with 
        the result.
        
        The units of the boundary work are kW.
        """
        
        def Wdot_one_CV(CVindex):
            """ calculate the p-v work for one CV """
            
            x0_raw = self.t[0:self.Ntheta]
            y0_raw = self.p[CVindex, 0:self.Ntheta]*self.dV[CVindex, 0:self.Ntheta]
            
            # Convert into chunks that are delimited by nan, if any
            isnotnan_indices = np.flatnonzero(~np.isnan(y0_raw))
            breaks = np.flatnonzero(np.diff(isnotnan_indices) > 1)

            if len(breaks) != 0:
                chunks = np.split(isnotnan_indices, np.array(breaks)+1)
            else:
                chunks = [isnotnan_indices]
                
            return -sum([trapz(y0_raw[ii], x0_raw[ii]) for ii in chunks])*self.omega/(2*pi)
            
        self.Wdot_pv = 0.0
        for CVindex in range(self.p.shape[0]):
            self.Wdot_pv+=Wdot_one_CV(CVindex)
        
    def post_cycle(self):
        """
        This stuff all happens at the end of the cycle.  It is a private method 
        not meant to be called externally
        
        The following things are done:
        
        #. The boundary work is calculated
        #. The flows are post-processed
        #. The heat transfer is post-processed
        #. The mass flow rate is calculated
        #. The volumetric efficiency is calculated
        #. The adiabatic efficiency is calculated
        #. The isentropic power is calculated
        #. The power input is calculated
        """
            
        self.calc_boundary_work()
        self._postprocess_flows()
        self._postprocess_HT()
        
        # Calculate the lumped mass energy balance
        if self.callbacks.lumps_energy_balance_callback is not None:
            self.lumps_resid = self.callbacks.lumps_energy_balance_callback()
            #  Convert to an arraym if needed
            if not isinstance(self.lumps_resid, arraym):
                self.lumps_resid = arraym(self.lumps_resid)
        else:
            raise ValueError('lumps_energy_balance_callback cannot be None')
        
        if not hasattr(self,'Qamb'):
            self.Qamb = 0
        
        # The total mass flow rate
        self.mdot = self.FlowsProcessed.mean_mdot[self.key_inlet]
        
        for key, State in six.iteritems(self.Tubes.Nodes):
            if key == self.key_inlet:
                 inletState = State
            if key == self.key_outlet:
                 outletState = State

        try:
            Vdisp = self.Vdisp

        except:
            Vdisp = self.Vdisp()

        self.eta_v = self.mdot / (self.omega/(2*pi)*Vdisp*inletState.rho)

        h1 = inletState.h
        h2 = outletState.h
        s1 = inletState.s

        # Can't use intermediate temperature because the state might be two-phase
        # for some conditions and you are better off just calculating the enthalpy
        # directly
        temp = outletState.copy()
        temp.update(dict(P=outletState.p, S=s1))
        h2s = temp.h
        
        if outletState.p > inletState.p:
            # Compressor Mode
            self.eta_a = (h2s-h1)/(h2-h1)
            self.Wdot_i = self.mdot*(h2s-h1)
        else:
            # Expander Mode
            self.eta_a = (h1-h2)/(h1-h2s)
            self.Wdot_i = self.mdot*(h1-h2s) 
            
        # self.Qamb is positive if heat is being added to the lumped mass
        self.Wdot = self.mdot*(h2-h1)-self.Qamb
    
    def _check_cycle_abort(self, index, I = 100):
        """
        This function will check whether an abort has been requested every 
        ``I`` steps of the solver throughout the rotation
        
        Meant for calling by cycle_RK45, cycle_SimpleEuler, cycle_Heun, etc.
        
        Primarily this is useful for use with the GUI, where the GUI can pass
        an abort command to the model
        
        Parameters
        ----------
        index : int
            The index of the step
        I : int, optional
            Check abort at this interval
        
        """
        # % is the symbol for modulus in python
        if index % I == 0 and self.Abort():
            self._want_abort = True
            return True
        
    def check_abort(self):
        """
        A callback for use with the graphical user interface to force solver to quit
        
        It will check the Scroll.pipe_abort pipe for a ``True`` value, and if it
        finds one, it will set the Scroll._want_abort value to ``True`` which 
        will be read by the main execution thread
        
        Once ``self._want_abort`` is ``True``, it will stay latched ``True`` until the 
        run is terminated
        """
        
        #  If you received an abort request, set a flag in the simulation
        if self.pipe_abort.poll() and self.pipe_abort.recv():
            print('received an abort request')
            self._want_abort = True
        
        #  If the run has timed out, quit
        if clock() - self.start_time > self.timeout:
            print('run timed out')
            self._want_abort = True
            
        return self._want_abort
        
    def precond_solve(self,**kwargs):
        """
        This function is deprecated and will be removed in a future version
        """
            
        import warnings
        msg = 'precond_solve is deprecated and will be removed in a future version. Please use solve instead'
        warnings.warn(msg, DeprecationWarning)
        self.solve(**kwargs)
       
    def connect_callbacks(self,
                          step_callback=None,
                          heat_transfer_callback=None,
                          lumps_energy_balance_callback=None,
                          endcycle_callback=None
                          ):
        """ 
        Connect up the callbacks for the simulation
        
        The callbacks must either be unbound methods or methods of a class derived from PDSimCore
        
        No keyword arguments are supported to be passed to the callbacks.  The 
        callback is probably a bound method of a PDSimCore instance, in which 
        case you have access to all the data in the class anyway
        
        Parameters
        ----------
        step_callback : function, or :class:`StepCallback <PDSim.core.callbacks.StepCallback>` subclass
            
            If a function is provided, it must have the call signature::
            
                disable_adaptive,h = step_callback(double t, double h, int i)
                
            where ``h`` is the step size that the adaptive solver wants to use, ``t`` is the current value of the independent variable, and ``i`` is the index in the container variables.  The return value ``disableAdaptive`` is a boolean value that describes whether the adaptive method should be turned off for this step ( ``False`` : use the adaptive method), and ``h`` is the step size you want to use.  If you don't want to disable the adaptive method and use the given step size, just::
                
                return False,h
            
            in your code.
        
        heat_transfer_callback : function, or :class:`HeatTransferCallback <PDSim.core.callbacks.HeatTransferCallback>` subclass
            
            If a function is provided, the heat_transfer_callback function must have the call signature::
            
                Q = heat_transfer_callback(double t)
            
            It should return an :class:`arraym <PDSim.misc.datatypes.arraym>` instance 
            with the same length as the number of CV in existence.  
            The entry in the :class:`arraym <PDSim.misc.datatypes.arraym>` is 
            positive if the heat transfer is TO the fluid in the CV in order 
            to maintain the sign convention that energy (or mass) input is 
            positive.
            
        lumps_energy_balance_callback : function, or :class:`LumpsEnergyBalanceCallback <PDSim.core.callbacks.LumpsEnergyBalanceCallback>` subclass
            
            If a function is provided, the lumps_energy_balance_callback 
            function must have the call signature::
            
                r = lumps_energy_balance_callback()
            
            It should return an :class:`arraym <PDSim.misc.datatypes.arraym>` 
            instance with the same length as the number of lumps.  The entry in 
            ``r`` is the value of the energy balance.  It will be driven to zero 
            by the solver
            
        """
        
        if step_callback is None:
            #No callback is provided, don't do anything
            pass
        elif isinstance(step_callback, PDSim.core.callbacks.StepCallback):
            #If the cythonized step callback is provided, hold onto it
            self.callbacks.step_callback = step_callback
        #Otherwise, wrap the desired callback if it has the right signature
        else:
            #Check the functional call
            callargs = inspect.getcallargs(step_callback, 0.0, 1e-10, 0)
            
            # Either a non-bound method is provided, or bound method is provided, in which case you get self,t,h,i as the values
            # t is a subclass of float, h is a subclass of float, is a subclass of int, and self is subclass of PDSimCore
            if not all([isinstance(arg,(float,int,PDSimCore)) for arg in callargs.values()]):
                sig_ok = False
            else:
                if len(callargs) in [3,4]:
                    sig_ok = True
                else:
                    sig_ok = False
            
            if step_callback is not None and sig_ok:
                self.callbacks.step_callback = PDSim.core.callbacks.WrappedStepCallback(self, step_callback)
            else:
                raise ValueError("step_callback is not possible to be wrapped - neither a subclass of StepCallback nor acceptable function signature")
        
        if heat_transfer_callback is None:
            #No callback is provided, don't do anything
            pass
        elif isinstance(heat_transfer_callback, PDSim.core.callbacks.HeatTransferCallback):
            #If the cythonized heat transfer callback is provided, hold a pointer to it
            self.callbacks.heat_transfer_callback = heat_transfer_callback
        else:
            callargs = inspect.getcallargs(heat_transfer_callback, 0.0)
            # Either a non-bound method is provided, or bound method is provided, in which case you get self,t as the values
            # t is a subclass of float, and self is subclass of PDSimCore
            if not all([isinstance(arg,(float,int,PDSimCore)) for arg in callargs.values()]):
                sig_ok = False
            else:
                if len(callargs) in [1,2]:
                    sig_ok = True
                else:
                    sig_ok = False
            
            #Otherwise, wrap the desired callback if it has the right signature
            if heat_transfer_callback is not None and sig_ok:
                self.callbacks.heat_transfer_callback = PDSim.core.callbacks.WrappedHeatTransferCallback(self, heat_transfer_callback)
            else:
                raise ValueError("heat_transfer_callback is not possible to be wrapped - neither a subclass of HeatTransferCallback nor an acceptable function")
        
        if lumps_energy_balance_callback is None:
            #No callback is provided, don't do anything
            pass
        elif isinstance(lumps_energy_balance_callback, PDSim.core.callbacks.LumpsEnergyBalanceCallback):
            #If the cythonized lump energy balance callback is provided, hold onto it
            self.callbacks.lumps_energy_balance_callback = lumps_energy_balance_callback
        #Otherwise, wrap the desired callback if it has the right signature
        else:
            callargs = inspect.getcallargs(lumps_energy_balance_callback)
            # Either a non-bound method is provided, or bound method is provided, in which case you get self,t as the values
            # t is a subclass of float, and self is subclass of PDSimCore
            sig_ok = len(callargs) == 0 or (len(callargs) == 1 and isinstance(list(callargs.values())[0],PDSimCore))
            
            if lumps_energy_balance_callback is not None and sig_ok:  #Do functional introspection here where the ``True`` is
                self.callbacks.lumps_energy_balance_callback = PDSim.core.callbacks.WrappedLumpsEnergyBalanceCallback(self, lumps_energy_balance_callback)
            else:
                raise ValueError("lump_energy_balance_callback is not possible to be wrapped - neither a subclass of LumpsEnergyBalanceCallback nor an acceptable function")
        
        self.callbacks.endcycle_callback = endcycle_callback
    
    def one_cycle(self, 
                  X, 
                  cycle_integrator = 'RK45',
                  cycle_integrator_options = None):
        """
        Only run one cycle
        
        Parameters
        ----------
        cycle_integrator : string
            One of 'RK45','Euler','Heun'
        cycle_integrator_options : dictionary
            options to be passed to the solver function (RK45, Euler, etc.)
        """
        # Make cycle_integrator_options an empty dictionary if not provided
        if cycle_integrator_options is None:
            cycle_integrator_options = {}
            tmax = 2*math.pi
        else:
            tmax = cycle_integrator_options['tmax']
                
        X = arraym(X)
                
        # (1). First, run all the tubes
        for tube in self.Tubes:
            tube.TubeFcn(tube)
            
        # Call update_existence to save the enthalpies for the tubes 
        self.update_existence()
        
        try:
            t1 = clock()
            # Run the pre-cycle code
            self.pre_cycle()

            if cycle_integrator == 'Euler':
                # Default to 7000 steps if not provided
                N = getattr(self,'EulerN', 7000)                
                integrator = EulerIntegrator(self, X)
                aborted = integrator.do_integration(N, 0, tmax)
            elif cycle_integrator == 'Heun':
                # Default to 7000 steps if not provided
                N = getattr(self,'HeunN', 7000)
                integrator = HeunIntegrator(self, X)
                aborted = integrator.do_integration(N, 0, tmax)
            elif cycle_integrator == 'RK45':
                # Default to tolerance of 1e-8 if not provided
                eps_allowed = getattr(self,'RK45_eps', 1e-8)
                integrator = RK45Integrator(self, X)
                aborted = integrator.do_integration(0, tmax, eps_allowed=eps_allowed)
            else:
                raise AttributeError('solver_method should be one of RK45, Euler, or Heun')
            
            if aborted == False:
                integrator.post_integration()
                
                self.Itheta = integrator.Itheta
                self.Ntheta = self.Itheta + 1

                # Make sure we got the right number of things
                assert self.Ntheta == len(self.FlowStorage)
                self.post_cycle()
            
        except ValueError as VE:
            # debug_plots(self)
            raise
        
        if aborted is None:
            aborted = False
                    
        #  Quit if you have aborted in one of the cycle solvers
        if aborted == 'abort':
            return None
        
        t2 = clock()
        print('Elapsed time for cycle is {0:g} s'.format(t2-t1))
        
        mdot_out = self.FlowsProcessed.mean_mdot[self.key_outlet]
        mdot_in = self.FlowsProcessed.mean_mdot[self.key_inlet]
        
        # We need to find the key at the inlet to the outlet tube.
        Tube = self.Tubes[self.key_outlet]
        if Tube.key1 == self.key_outlet:
            key_outtube_inlet = Tube.key2
        elif Tube.key2 == self.key_outlet:
            key_outtube_inlet = Tube.key1
            
        # This is the so-called hd' state at the outlet of the pump set
        self.h_outlet_pump_set = (self.FlowsProcessed.integrated_mdoth[key_outtube_inlet]
                                 /self.FlowsProcessed.integrated_mdot[key_outtube_inlet])
        
        # It should be equal to the enthalpy of the fluid at the inlet
        # to the outlet tube at the current Td value
        h_outlet_Tube = self.Tubes.Nodes[key_outtube_inlet].h
        # Residual is the difference of these two terms
        # We put it in kW by multiplying by flow rate
        self.resid_Td = 0.1*(h_outlet_Tube - self.h_outlet_pump_set)
        
    def OBJECTIVE_CYCLE(self, Td_Tlumps0, X, epsilon = 0.003, cycle_integrator = 'RK45', OneCycle = False, cycle_integrator_options = None, plot_every_cycle = False):
        """
        The Objective function for the energy balance solver
        
        Parameters
        ----------
        Td_Tlumps0 : list
            Discharge temperature and lump temperatures
        X : :class:`arraym <PDSim.misc.datatypes.arraym>` instance
            Contains the state variables for all the control volumes in existence, as well as any other integration variables
        epsilon : float
            Convergence criterion applied to all of the solvers
        cycle_integrator : string, one of 'RK45','Euler','Heun'
            Which solver is to be used to integrate the steps
        OneCycle : boolean
            If ``True``, stop after one cycle
        plot_every_cycle : boolean
            If ``True``, make the debug plots at every cycle
        cycle_integrator_options : dictionary
            Options to be passed to cycle integrator
        """
        
        # Consume the first element as the discharge temp 
        self.Td = float(Td_Tlumps0.pop(0))
        # The rest are the lumps in order
        self.Tlumps = Td_Tlumps0   
        
        #  The first time this function is run, save the state variables
        if self.xstate_init is None:
            self.xstate_init = X
            self.exists_CV_init = self.CVs.exists_keys
        
        i = 0
        worst_error = 1000
        while worst_error > epsilon:
            
            #  Actually run the cycle, runs post_cycle at the end,
            #  sets the parameter lumps_resid in this class
            #  Also sets resid_Td
            self.one_cycle(X, 
                           cycle_integrator = cycle_integrator,
                           cycle_integrator_options = cycle_integrator_options)
            
            if self.Abort():
                return
                
            errors, X = self.callbacks.endcycle_callback()
            error_metric = np.sqrt(np.sum(np.power(errors, 2)))
            
            ###  -----------------------------------
            ###         The lump temperatures
            ###  -----------------------------------
                
            self.solvers.lump_eb_history.append([self.Tlumps, self.lumps_resid])
            
            if len(self.solvers.lump_eb_history) == 1:
                
                T, EB = self.solvers.lump_eb_history[-1]
                
                #  T and EB are one-element lists, get floats
                _T, _EB = T[0], EB[0]
            
                # Use the thermal mass to make the step
                # Here is the logic:
                # Instantaneous energy balance given by
                # dU/dt = m*c*(dT/dt) = sum(Qdot)
                # and if dt = one cycle period (seconds/rev) DeltaT = 2*pi/omega
                # DELTAT = sum(Qdot)*Deltat/(m*c)
                thermal_capacitance = 0.49*0.001 # [kJ/K]
                Deltat = (2*np.pi)/self.omega # [s]
            
                #  Update the lump temperatures
                self.Tlumps = [_T + _EB*Deltat/thermal_capacitance]
            
            else:
                
                #  Get the values from the history
                Tn1, EBn1 = self.solvers.lump_eb_history[-1]
                Tn2, EBn2 = self.solvers.lump_eb_history[-2]
                
                #  Convert to numpy arrays
                Tn1, EBn1, Tn2, EBn2 = [np.array(l) for l in [Tn1, EBn1, Tn2, EBn2]]
                
                #  Use the relaxed secant method to find the solution 
                Tnew = Tn1 - 0.7*EBn1*(Tn1-Tn2)/(EBn1-EBn2)
            
                #  Update the lump temperatures    
                self.Tlumps = Tnew.tolist()
            
            ###  -----------------------------------
            ###        The discharge enthalpy
            ###  -----------------------------------
            
            #  The outlet tube
            outlet_tube = self.Tubes[self.key_outlet]
            
            #  Get the keys for the elements of the outlet tube
            if outlet_tube.key1 == self.key_outlet:
                key_outtube_inlet = outlet_tube.key2
                key_outtube_outlet = outlet_tube.key1
            elif outlet_tube.key2 == self.key_outlet:
                key_outtube_inlet = outlet_tube.key1
                key_outtube_outlet = outlet_tube.key2
                
            if error_metric < 0.1*epsilon and abs(self.lumps_resid[0]) < epsilon:

                # Each time that we get here and we are significantly below the threshold, store the values
            
                #  Get the current value for the outlet enthalpy of the machine
                h_outlet = self.Tubes[self.key_outlet].State2.get_h()
                    
                # Store the values in the list of values
                self.solvers.hdisc_history.append([h_outlet,self.resid_Td])

                if len(self.solvers.hdisc_history) == 1:
                    # The first time we get here, perturb the discharge enthalpy
                    self.Tubes.Nodes[self.key_outlet].update_ph(self.Tubes.Nodes[self.key_outlet].p, h_outlet + 5)
                else:
                    #  Get the values from the history
                    hdn1, EBn1 = self.solvers.hdisc_history[-1]
                    hdn2, EBn2 = self.solvers.hdisc_history[-2]
                    
                    #  Use the relaxed secant method to find the solution 
                    hdnew = hdn1 - 0.75*EBn1*(hdn1-hdn2)/(EBn1-EBn2)
                    
                    #  Reset the outlet enthalpy of the outlet tube based on our new
                    #  value for it
                    self.Tubes.Nodes[self.key_outlet].update_ph(self.Tubes.Nodes[self.key_outlet].p, hdnew)

                print(self.solvers.hdisc_history)
                
            print('new outlet T', self.Tubes.Nodes[self.key_outlet].T)
                
            # Store a copy of the initial temperatures of the chambers
            self.solvers.initial_states_history.append(self.T[:,0].copy())
                
            if OneCycle:
                print('Quitting due to OneCycle being set to True')
                return
                
            if plot_every_cycle:
                from PDSim.plot.plots import debug_plots
                debug_plots(self)
            
            if self.Abort():
                print('Quitting because Abort flag hit')
                return
            
            #  Reset the flag for the fixed side of the outlet tube
            #outlet_tube.fixed = old_fixed
            
            mdot_out = abs(self.FlowsProcessed.mean_mdot[self.key_outlet])
            mdot_in = abs(self.FlowsProcessed.mean_mdot[self.key_inlet])
            mdot_error = (mdot_out/mdot_in-1)*100
            
            print('===========')
            print('|| # {i:03d} ||'.format(i=i))
            print('===========')
            print(error_ascii_bar(abs(self.lumps_resid[0]), epsilon), 'energy balance kW ', self.lumps_resid[0], ' Tlumps: ',self.Tlumps,'K')
            print(error_ascii_bar(abs(self.resid_Td), epsilon), 'discharge state', self.resid_Td, 'h_pump_set: ', self.h_outlet_pump_set,'kJ/kg', self.Tubes.Nodes[key_outtube_inlet].h, 'kJ/kg')
            print(error_ascii_bar(error_metric, epsilon), 'cycle-cycle    ', error_metric)
            print(error_ascii_bar(abs(mdot_error), 1), 'mdot [%]', mdot_error, '|| in:', mdot_in*1000, 'g/s || out:', mdot_out*1000, 'g/s ')
            
            worst_error = max(abs(self.lumps_resid[0]), abs(self.resid_Td), np.sqrt(np.sum(np.power(errors, 2))))
            i += 1
        
        #  If the abort function returns true, quit this loop
        if self.Abort():
            print('Quitting OBJECTIVE_CYCLE loop in core.solve')
            return None #  Stop
        else:
            if len(self.solvers.hdisc_history) == 0:
                # Store the values in the list of values
                self.solvers.hdisc_history.append([self.Tubes[self.key_outlet].State2.get_h(),self.resid_Td])
            
    def solve(self,
              key_inlet = None,
              key_outlet = None,
              solver_method = 'Euler',
              OneCycle = False,
              Abort = None,
              pipe_abort = None,
              UseNR = False,
              alpha = 0.5,
              plot_every_cycle = False,
              x0 = None,
              reset_initial_state = False,
              timeout = 3600,
              eps_cycle = 0.001,
              eps_energy_balance = 0.01,
              cycle_integrator_options = None,
              max_number_of_steps = 40000,
              **kwargs):
        """
        This is the driving function for the PDSim model.  It can be extended through the 
        use of the callback functions
        
        It is highly recommended to call this function using keyword arguments like::
        
            solve(key_inlet = 'inlet.1', 
                  key_outlet = 'outlet.1', ....)
        
        Parameters
        ----------
        key_inlet : string
            The key for the flow node that represents the upstream quasi-steady point
        key_outlet : string
            The key for the flow node that represents the upstream quasi-steady point
        solver_method : string
        OneCycle : boolean
            If ``True``, stop after just one rotation.  Useful primarily for 
            debugging purposes
        Abort : function
            A function that may be called to determine whether to stop running.  
            If calling Abort() returns ``True``, stop running 
        pipe_abort : 
        UseNR : boolean
            If ``True``, use a multi-dimensional solver to determine the initial state of the state variables for each control volume
        alpha : float
            Use a range of ``(1-alpha)*dx, (1+alpha)*dx`` for line search if needed
        plot_every_cycle : boolean
            If ``True``, make the plots after every cycle (primarily for debug purposes)
        x0 : arraym
            The starting values for the solver that modifies the discharge temperature and lump temperatures
        reset_initial_state : boolean
            If ``True``, use the stored initial state from the previous call to ``solve`` as the starting value for the thermodynamic values for the control volumes
        timeout : float
            Number of seconds before the run times out
        eps_cycle : float
            Cycle-cycle convergence criterion
        eps_energy_balance : float
            Energy balance convergence criterion
        cycle_integrator_options : dictionary
            A dictionary of options to be passed to the cycle integrator
        max_number_of_steps : int
            Maximum number of steps allowed per rotation
        
        Notes
        -----
        The callbacks ``step_callback`` and ``endcycle_callback`` and 
        ``heat_transfer_callback`` and ``lump_energy_balance_callback`` and 
        ``valves_callback`` should now be passed to the connect_callbacks() 
        function before running precond_solve() or solve()
        
        """
        if any(cb in kwargs for cb in ['step_callback','endcycle_callback','heat_transfer_callback','lump_energy_balance_callback','valves_callback']):
            raise NotImplementedError('callback functions are no longer passed to solve() function, rather they are passed to connect_callbacks() function prior to calling solve()')
        
        #  Save copies of the inlet and outlet states at the root of the HDF5 file
        #  for ease of retrieval
        self.inlet_state = self.Tubes.Nodes[key_inlet] 
        self.outlet_state = self.Tubes.Nodes[key_outlet]
        
        # Carry out some pre-run checks
        self._check()
        
        from time import clock
        
        self.start_time = clock()
        self.timeout = timeout
        
        #Connect functions that have been serialized by saving the function name as a string
        self.connect_flow_functions()
        
        #Both inlet and outlet keys must be connected to invariant nodes - 
        # that is they must be part of the tubes which are all quasi-steady
        if not key_inlet == None and not key_inlet in self.Tubes.Nodes:
            raise KeyError('key_inlet must be a Tube node')
        if not key_outlet == None and not key_outlet in self.Tubes.Nodes:
            raise KeyError('key_outlet must be a Tube node')
        
        self.key_inlet = key_inlet
        self.key_outlet = key_outlet
        
        from time import clock
        t1=clock()
        
        # Set up a pipe for accepting a value of True which will abort the run
        # Used from the GUI to kill process from the top-level thread
        self.pipe_abort = pipe_abort
        
        if len(self.CVs) < 1:
            raise ValueError('At least one control volume must be added using the add_CV function')
            
        if len(self.Flows) <= 1:
            raise ValueError('At least two flows must be added using the add_flow function')
        
        # If a function called pre_solve is provided, call it with no input arguments
        if hasattr(self,'pre_solve'):
            self.pre_solve()
            
        # This runs before the model starts at all
        self.pre_run(N = max_number_of_steps)
        
        # Check which method is used to do aborting
        if Abort is None and pipe_abort is not None:
            # Use the pipe_abort pipe to look at the abort pipe to see whether 
            # to quit 
            self.Abort = self.check_abort
        elif Abort is None and pipe_abort is None:
            #Disable the ability to abort, always don't abort
            self.Abort = lambda : False
        elif Abort is not None and pipe_abort is None:
            self.Abort = Abort
        else:
            raise ValueError('Only one of Abort and pipe_abort may be provided')
        
        # If you want to reset the initial state, use the values that were
        # cached in the xstate_init array
        if reset_initial_state is not None and reset_initial_state:
            self.reset_initial_state()
            self.pre_cycle(self.xstate_init)
        else:
            # (2) Run a cycle with the given values for the temperatures
            self.pre_cycle()
            #x_state only includes the values for the chambers, the valves start closed
            #Since indexed, yields a copy
            self.x_state = self._get_from_matrices(0)[0:len(self.stateVariables)*self.CVs.Nexist]
        
        if x0 is None:
            x0 = [self.Tubes.Nodes[key_outlet].T, self.Tubes.Nodes[key_outlet].T]

        #  Actually run the solver
        self.OBJECTIVE_CYCLE(x0, self.x_state, 
                             cycle_integrator = solver_method, 
                             OneCycle = OneCycle,
                             epsilon = eps_energy_balance,
                             cycle_integrator_options = cycle_integrator_options,
                             plot_every_cycle = plot_every_cycle
                             ) 
                
        if not self.Abort() and not OneCycle: 
            self.post_solve()
            
        if hasattr(self,'resid_Td'):
            del self.resid_Td
            
        #  Save the elapsed time for simulation
        self.elapsed_time = clock() - t1
        
    def get_prune_keys(self):
        """
        Remove some elements when the simulation finishes that are not 
        very useful and/or are very large when stored to file
        
        Returns
        -------
        prune_key_list: list
            A list of HDF5 keys that are to be removed from the HDF5 file.
        """
        
        return ['/CVs/CVs',
                '/CVs/Nodes',
                '/CVs/T',
                '/CVs/cp',
                '/CVs/cv',
                '/CVs/dpdT',
                '/CVs/exists_CV',
                '/CVs/exists_indices',
                '/CVs/exists_keys',
                '/CVs/h',
                '/CVs/p',
                '/CVs/rho',
                '/callbacks',
                '/core',
                '/steps',
                '/theta',
                '/Flows'
                ]
        
    def attach_HDF5_annotations(self, fName):
        """
        In this function, annotations can be attached to each HDF5 field
        
        Parameters
        ----------
        fName : string
            The file name for the HDF5 file that is to be used
        """ 
        attrs_dict = {
                '/t':'The array of the independent variable in the solution, either time or crank angle [rad or s]',
                '/m':'The NCV x Nt matrix with the mass in each control volume [kg]',
                '/T':'The NCV x Nt matrix with the temperature in each control volume [K]',
                '/V':'The NCV x Nt matrix with the volume in each control volume [m^3]',
                '/dV':'The NCV x Nt matrix with the derivative of volume w.r.t. crank angle in each control volume [m^3/radian]',
                '/h':'The NCV x Nt matrix with the enthalpy in each control volume [kJ/kg]',
                '/p':'The NCV x Nt matrix with the pressure in each control volume [kPa]',
                '/rho':'The NCV x Nt matrix with the density in each control volume [kg/m^3]',
                '/Q':'The NCV x Nt matrix with the heat transfer TO the gas in each control volume [kW]',
                '/xL':'The NCV x Nt matrix with the oil mass fraction in each control volume [-]',
                '/A_shell':'The shell area of the machine [m^2]',
                '/h_shell':'The heat transfer coefficient between the shell and the ambient [kW/m^2/K]',
                '/key_inlet':'The key for the inlet node',
                '/key_outlet':'The key for the outlet node',
                '/elapsed_time':'The elapsed time for the simulation run [s]',
                '/eta_a': 'Adiabatic efficiency [-]',
                '/eta_oi':'Overall isentropic efficiency [-]',
                '/eta_v':'Volumetric efficiency [-]',
                '/Qamb':'Rate of heat transfer from the machine TO the ambient [kW]',
                '/RK45_eps':'Step error tolerance for Runge-Kutta method [varied]',
                '/Tamb':'Ambient temperature [K]',
                '/Wdot_pv':'Mechanical power calculated as the integral of -pdV [kW]',
                '/Wdot_electrical':'Electrical power of the machine [kW]',
                '/Wdot_forces':'Mechanical power calculated from the mechanical analysis [kW]',
                '/mdot':'Mass flow rate [kg/s]',
                '/motor/eta_motor':'Motor efficiency [-]',
                '/motor/losses':'Losses generated in the motor [kW]',
                '/motor/suction_fraction':'Fraction of the motor losses that are added to the suction gas [-]',
                '/motor/type':'The model used to simulate the motor',
                '/omega':'Rotational speed [rad/s]',
                '/run_index':'A unique identifier for runs in a batch'
                }
        
        hf = h5py.File(fName,'a')
        
        for k, v in attrs_dict.iteritems():
            dataset = hf.get(k)
            if dataset is None:
                print('bad key',k)
            else:
                dataset.attrs['note'] = v
        hf.close()
        
    def post_solve(self):
        """
        Do some post-processing to calculate flow rates, efficiencies, etc.  
        """      

        #Resize all the matrices to keep only the real data
        print('Ntheta is', self.Ntheta)
        self.t = self.t[  0:self.Ntheta]
        self.T = self.T[:,0:self.Ntheta]
        self.p = self.p[:,0:self.Ntheta]
        self.Q = self.Q[:,0:self.Ntheta]
        self.m = self.m[:,0:self.Ntheta]
        self.rho = self.rho[:,0:self.Ntheta]
        self.V = self.V[:,0:self.Ntheta]
        self.dV = self.dV[:,0:self.Ntheta]
        self.h = self.h[:,0:self.Ntheta]
        self.xValves = self.xValves[:,0:self.Ntheta]
        
        print('mdot*(h2-h1),P-v,Qamb', self.Wdot, self.Wdot_pv, self.Qamb)
        print('Mass flow rate is',self.mdot*1000,'g/s')
        print('Volumetric efficiency is',self.eta_v*100,'%')
        print('Adiabatic efficiency is',self.eta_a*100,'%')
        
        # Restructure the history for easier writing to file and more clear description of what the things are
        hdisc_history = list(zip(*self.solvers.hdisc_history))
        self.solvers.hdisc_history = dict(hd = np.array(hdisc_history[0]), hd_error = np.array(hdisc_history[1]))
        lump_eb_history = list(zip(*self.solvers.lump_eb_history))
        self.solvers.lump_eb_history = dict(Tlumps = np.array(lump_eb_history[0]), lump_eb_error = np.array(lump_eb_history[1]))
        self.solvers.initial_states_history = np.array(zip(*self.solvers.initial_states_history)).T

        
    def derivs(self, theta, x):
        """
        Evaluate the derivatives of the state variables
        
        derivs() is an internal function that should (probably) not actually be called by
        any user-provided code, but is documented here for completeness.
        
        Parameters
        ----------
        theta : float
            The value of the independent variable
        x : :class:`arraym <PDSim.misc.datatypes.arraym>` instance
            The array of the independent variables (state variables plus valve parameters)
        
        Returns
        -------
        dfdt : :class:`arraym <PDSim.misc.datatypes.arraym>` instance
        
        """

        # Updates the state, calculates the volumes, prepares all the things needed for derivatives
        self.core.properties_and_volumes(self.CVs.exists_CV, theta, STATE_VARS_TM, x)
        
        # Calculate the flows and sum up all the terms
        self.core.calculate_flows(self.Flows)
        
        # Calculate the heat transfer terms if provided
        if self.callbacks.heat_transfer_callback is not None:
            self.core.Q = arraym(self.callbacks.heat_transfer_callback(theta))
            if not len(self.core.Q) == self.CVs.Nexist:
                raise ValueError('Length of Q is not equal to length of number of CV in existence')
        else:
            self.core.Q = empty_arraym(self.CVs.Nexist)
            
        # Calculate the derivative terms and set the derivative of the state vector
        self.core.calculate_derivs(self.omega, False)
        
        # Add the derivatives for the valves
        if self.__hasValves__:
            # 
            offset = len(self.stateVariables)*self.CVs.Nexist
            for i, valve in enumerate(self.Valves):
                if x[offset+2+i*2-1] < -10:
                    x[offset+2+i*2-1]=0.0
                if x[offset+i*2] > valve.x_stopper and x[offset+2+i*2-1] > 0.0 :
                    x[offset+2+i*2-1]=0.0 
                # Get the values from the input array for this valve
                xvalve = x[offset+i*2:offset+2+i*2]
                # Set the values in the valve class
                valve.set_xv(xvalve)
                # Get the derivatives of position and derivative of velocity                
                self.core.property_derivs.extend(valve.derivs(self))
        
        return self.core.property_derivs
        
    def valves_callback(self):
        """
        This is the default valves_callback function that builds the list of 
        derivatives of position and velocity with respect to the crank angle
        
        It returns a :class:`list` instance with the valve return values in order 
        """
        #Run each valve model in turn to calculate the derivatives of the valve parameters
        # for each valve
        f=[]
        for Valve in self.Valves:
            f+=Valve.derivs(self)
        return f
        
    def IsentropicNozzleFM(self,FlowPath,A,**kwargs):
        """
        A generic isentropic nozzle flow model wrapper
        
        Parameters
        ----------
        FlowPath : :class:`FlowPath <PDSim.flow.flow.FlowPath>` instance
            A fully-instantiated flow path model
        A : float
            throat area for isentropic nozzle model [:math:`m^2`]
            
        Returns
        -------
        mdot : float
            The mass flow through the flow path [kg/s]
        """

        try:
            mdot = flow_models.IsentropicNozzle(A,
                                                FlowPath.State_up,
                                                FlowPath.State_down)
            return mdot
        except ZeroDivisionError:
            return 0.0
        
    def IsentropicNozzleFMSafe(self,FlowPath,A,DP_floor,**kwargs):
        """
        A generic isentropic nozzle flow model wrapper with the added consideration
        that if the pressure drop is below the floor value, there is no flow.
        This was added to handle the case of the injection line where there is
        no flow out of the injection which greatly increases the numerical 
        stiffness 
        
        Parameters
        ----------
        FlowPath : :class:`FlowPath <PDSim.flow.flow.FlowPath>`
            A fully-instantiated flow path model
        A : float
            throat area for isentropic nozzle model [:math:`m^2`]
        DP_floor: float
            The minimum pressure drop [kPa]
            
        Returns
        -------
        mdot : float
            The mass flow through the flow path [kg/s]
        """

        try:
            if FlowPath.State_up.p-FlowPath.State_down.p > DP_floor:
                mdot = flow_models.IsentropicNozzle(A,
                                                    FlowPath.State_up,
                                                    FlowPath.State_down)
                return mdot
            else:
                return 0.0
        except ZeroDivisionError:
            return 0.0

    def step_callback(self,t,h,i):
        """ The default step_callback which does nothing
        
        Parameters
        ----------
        t : float
            The current value of the independent variable
        h : float
            The current step size
        i : int
            The current index
        """
        return False,h
        
    def endcycle_callback(self, eps_wrap_allowed=0.0001):
        """
        This function can be called at the end of the cycle if so desired.
        Its primary use is to determine whether the cycle has converged for a 
        given set of discharge temperatures and lump temperatures.
        
        Parameters
        ----------
        eps_wrap_allowed : float
            Maximum error allowed, in absolute value
        
        Returns 
        -------
        redo : boolean
            ``True`` if cycle should be run again with updated inputs, ``False`` otherwise.
            A return value of ``True`` means that convergence of the cycle has been achieved
        """
        assert self.Ntheta - 1 == self.Itheta
        #old and new CV keys
        LHS,RHS=[],[]
        errorT,error_rho,error_mass,newT,new_rho,new_mass,oldT,old_rho,old_mass={},{},{},{},{},{},{},{},{}
        
        for key in self.CVs.exists_keys:
            # Get the 'becomes' field.  If a list, parse each fork of list. If a single key convert 
            # into a list so you can use the same code below 
            if not isinstance(self.CVs[key].becomes, list):
                becomes = [self.CVs[key].becomes]
            else:
                becomes = self.CVs[key].becomes
                
            Iold = self.CVs.index(key)

            for newkey in becomes:
                #  If newkey is 'none', the control volume will die at the end
                #  of the cycle, so just keep going
                if newkey == 'none': continue
                Inew = self.CVs.index(newkey)
                newCV = self.CVs[newkey]
                # There can't be any overlap between keys
                if newkey in newT:
                    raise KeyError('newkey [' + newkey + '] is already in newT; becomes keys overlap but should not')
                #What the state variables were at the start of the rotation
                oldT[newkey]=self.T[Inew, 0]
                old_rho[newkey]=self.rho[Inew, 0]
                #What they are at the end of the rotation
                newT[newkey]=self.T[Iold,self.Itheta]
                new_rho[newkey]=self.rho[Iold,self.Itheta]
                errorT[newkey]=(oldT[newkey]-newT[newkey])/newT[newkey]
                error_rho[newkey]=(old_rho[newkey]-new_rho[newkey])/new_rho[newkey]
                #Update the list of keys for setting the exist flags
                LHS.append(key)
                RHS.append(newkey)
        
        error_T_list = [errorT[key] for key in self.CVs.keys if key in newT]
        error_rho_list = [error_rho[key] for key in self.CVs.keys if key in new_rho]
        
        new_T_list = [newT[key] for key in self.CVs.keys if key in newT]
        new_rho_list = [new_rho[key] for key in self.CVs.keys if key in new_rho]
        
        #Reset the exist flags for the CV - this should handle all the possibilities
        #Turn off the LHS CV
        for key in LHS:
            self.CVs[key].exists=False
        #Turn on the RHS CV
        for key in RHS:
            self.CVs[key].exists=True
            
        self.update_existence()

        # Error values are based on density and temperature independent of 
        # selection of state variables 
        error_list = []
        for var in ['T','D']:
            if var == 'T':
                error_list += error_T_list
            elif var == 'D':
                error_list += error_rho_list
            elif var == 'M':
                error_list += error_mass_list
            else:
                raise KeyError
            
        # Calculate the volumes at the beginning of the next rotation
        self.core.just_volumes(self.CVs.exists_CV, 0)
        V = {key:V for key,V in zip(self.CVs.exists_keys,self.core.V)}
        new_mass_list = [new_rho[key]*V[key] for key in self.CVs.exists_keys]
        
        new_list = []
        for var in self.stateVariables:
            if var == 'T':
                new_list += new_T_list
            elif var == 'D':
                new_list += new_rho_list
            elif var == 'M':
                new_list += new_mass_list
            else:
                raise KeyError
    
        return arraym(error_list), arraym(new_list)
    
    def connect_flow_functions(self):
        """
        Reconnect function pointers
        
        For pickling purposes, it can sometimes be useful to just give the name
        of the function relative to the PDSimCore (or derived class).  If the 
        function is just a string, reconnect it to the function in the PDSimCore 
        instance 
        """
        for Flow in self.Flows:
            if hasattr(Flow.MdotFcn, 'Function'):
                if isinstance(Flow.MdotFcn.Function, six.string_types):
                    if hasattr(self,Flow.MdotFcn.Function):
                        Flow.MdotFcn.Function = getattr(self, Flow.MdotFcn.Function)
                    else:   
                        raise AttributeError('The name of the function ['+Flow.MdotFcn.Function+']is not found in the PDSimCore derived class instance')
            
if __name__=='__main__':
    PC = PDSimCore()
    PC.attach_HDF5_annotations('runa.h5')
    print('This is the base class that is inherited by other compressor types.  Running this file doesn\'t do anything')



from CoolProp.State import State
from CoolProp.State cimport State

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.map cimport map
from cython.operator cimport dereference as deref, preincrement as inc
    
from PDSim.flow.flow_models import FlowFunction
from PDSim.flow.flow_models cimport FlowFunction

from PDSim.misc.datatypes import arraym
from PDSim.misc.datatypes cimport arraym

cdef class FlowPathCollection(list):
    
    
    cdef int N
    
    cdef int Nexists
    
    # The rotational speed [rad/s]
    cdef double omega
    
    cpdef update_existence(self, Core)
    cpdef calculate(self)
    cpdef get_deepcopy(self)
    cpdef sumterms(self, arraym summerdT, arraym summerdm)
    cpdef list flow_paths
        
#Make a stripped down class with the necessary terms included
cdef class FlowPath(object):
    
    cdef bytes m_key_up,m_key_down,m_key1,m_key2
    
    cdef public bint exists, key1_exists, key2_exists, key_up_exists, key_down_exists
    
    cdef public long key1Index, key2Index, key_up_Index, key_down_Index
    cdef public int ikey1, ikey2, ikey_up, ikey_down
    
    cdef public double mdot
    """ The mass flow rate [kg/s]"""
    
    cdef public double h_up
    """ The upstream enthalpy [kJ/kg] """ 
    
    cdef public double h_down
    """ The downstream enthalpy [kJ/kg] """
    
    cdef public double T_up
    """ The upstream temperature [K] """
    
    cdef public double p_up
    """ The upstream pressure [kPa] """
    
    cdef public double p_down
    """ The downstream pressure [kPa] """
    
    cdef public double A
    """ The flow area [m^2] """
    
    cdef public double edot
    """ The rate of irreversibility generation in this flow path [kW]"""
    
    cdef public FlowFunction MdotFcn
    """ The function that will return the mass flow rate """
    
    cdef public bytes MdotFcn_str
    
    cdef public State State1
    """ The first state """
    
    cdef public State State2
    """ The second state """
    
    cdef public State State_up
    """ The upstream state """
    
    cdef public State State_down
    """ The downstream state """
    
    cpdef dict __cdict__(self, AddStates = *)
    cpdef FlowPath get_deepcopy(self)
    cpdef calculate(self)



'''
Created on 2 oct. 2012 by maxime - Updated by Ian Bell

@author: ibell
Updated 2016/10/20 
@author: Davide Ziviani
'''
from __future__ import division
import sys
from PDSim.misc.clipper import pyclipper
import matplotlib.pyplot as plt
import numpy as np
 
clip = pyclipper.Pyclipper()

square = [[0,0], [1, 0], [1, 1], [0, 1], [0,0]]

x,y = zip(*square)

scale_factor = 10000000

#Rescale the dimensions to go from float to long
scaled_x = [_*scale_factor for _ in x]
scaled_y = [_*scale_factor for _ in y]
clip.subject_polygon([pair for pair in zip(scaled_x,scaled_y)])
plt.plot(x,y)

t = np.linspace(0,2*np.pi,1000)
x = 0.5+0.55*np.cos(t)
y = 0.5+0.55*np.sin(t)
scaled_x = [_*scale_factor for _ in x]
scaled_y = [_*scale_factor for _ in y]
clip.clip_polygon([pair for pair in zip(scaled_x,scaled_y)])
plt.plot(x,y)

sol = clip.execute(pyclipper.INTERSECTION)
for loop in sol:
    scaled_x,scaled_y = zip(*loop)
    x = [_/scale_factor for _ in scaled_x]
    y = [_/scale_factor for _ in scaled_y]
    plt.fill(x,y)

plt.show()


from __future__ import print_function

from math import pi
import numpy as np
from PDSim.scroll import scroll_geo
import PDSim.scroll.core as core
from PDSim.core.motor import Motor
from PDSim.flow.flow import FlowPath
from PDSim.flow.flow_models import IsentropicNozzleWrapper
from PDSim.core.containers import Tube,ControlVolume
import matplotlib.pyplot as plt
from CoolProp.CoolProp import Props
from CoolProp.State import State as CPState
from PDSim.core.core import PDSimCore
from PDSim.flow import flow_models
from PDSim.misc.datatypes import arraym

# If scipy is available, use its interpolation and optimization functions, otherwise, 
# use our implementation (for packaging purposes mostly)
try:
    import scipy.optimize as optimize
except ImportError:
    import PDSim.misc.solvers as optimize

class ScrollExpander(core.Scroll):
    """
    The expander is just like the compressor, except that the volumes go backwards.
    
    theta starts a revolution at 2*pi, and then decreases over one rotation to 0. Beta is equal to 2*pi-theta whc
    In this way, all the code for the compressor can be more readily used with the expander
    """
    
    def __init__(self, *args, **kwargs):
        core.Scroll.__init__(self, *args, **kwargs)
        
    def V_e1(self, beta, alpha=1, full_output=False):
        VdV = core.Scroll.V_c1(self, 2*pi-beta, alpha=alpha, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_e2(self, beta, alpha=1, full_output=False):
        VdV = core.Scroll.V_c2(self, 2*pi-beta, alpha=alpha, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_d1(self, beta):
        VdV = core.Scroll.V_s1(self, 2*pi-beta)
        return VdV[0],-VdV[1]
    
    def V_d2(self, beta):
        VdV = core.Scroll.V_s2(self, 2*pi-beta)
        return VdV[0],-VdV[1]
    
    def V_s1(self, beta, full_output=False):
        VdV = core.Scroll.V_d1(self, 2*pi-beta, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_s2(self, beta, full_output=False):
        VdV = core.Scroll.V_d2(self, 2*pi-beta, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_ss(self, beta, full_output=False):
        VdV = core.Scroll.V_dd(self, 2*pi-beta, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_sss(self, beta, full_output=False):
        if self.__before_discharge1__ == True and beta < 2*pi - self.theta_d:
            VdV = core.Scroll.V_dd(self, 2*pi-beta, full_output=full_output)
        else:
            VdV = core.Scroll.V_ddd(self, 2*pi-beta, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def V_da(self, beta, full_output=False):
        VdV = core.Scroll.V_sa(self, 2*pi-beta, full_output=full_output)
        return VdV[0],-VdV[1]
    
    def auto_add_CVs(self,inletState,outletState):
        """
        Adds all the control volumes for the scroll expander.
        
        Parameters
        ----------
        inletState
            A :class:`State <CoolProp.State.State>` instance for the inlet to the scroll set.  Can be approximate
        outletState
            A :class:`State <CoolProp.State.State>` instance for the outlet to the scroll set.  Can be approximate
            
        Notes
        -----
        Uses the indices of 
        
        ============= ===================================================================
        CV            Description
        ============= ===================================================================
        ``da``        Discharge Area
        ``d1``        Discharge chamber on side 1
        ``d2``        Discharge chamber on side 2
        ``s1``        Suction chamber on side 1
        ``s2``        Suction chamber on side 2
        ``ss``        Central suction chamber
        ``sss``       Merged suction chamber
        ``e1.i``      The i-th expansion chamber on side 1 (i=1 for outermost chamber)
        ``e2.i``      The i-th expansion chamber on side 2 (i=1 for outermost chamber)
        ============= ===================================================================
        """
        
        #Add all the control volumes that are easy.  Suction area and suction chambers
        self.add_CV(ControlVolume(key='da', initialState = outletState.copy(),
                                VdVFcn=self.V_da))
        self.add_CV(ControlVolume(key='d1', initialState = outletState.copy(),
                                VdVFcn=self.V_d1,becomes='none'))
        self.add_CV(ControlVolume(key='d2', initialState = outletState.copy(),
                                VdVFcn=self.V_d2,becomes='none'))
        
        #Discharge chambers are also easy.  Assume that you start with 'ddd' chamber merged.
        # No problem if this isn't true.
        self.add_CV(ControlVolume(key = 's1', initialState = inletState.copy(),
                                  VdVFcn = self.V_s1, exists = False, discharge_becomes = 'e1.'+str(scroll_geo.nC_Max(self.geo))))
        self.add_CV(ControlVolume(key = 's2', initialState = inletState.copy(),
                                  VdVFcn = self.V_s2, exists = False, discharge_becomes = 'e2.'+str(scroll_geo.nC_Max(self.geo))))
        self.add_CV(ControlVolume(key = 'ss', initialState = inletState.copy(),
                                  VdVFcn = self.V_ss, exists = False, discharge_becomes = 'sss'))
        self.add_CV(ControlVolume(key = 'sss', initialState = inletState.copy(),
                                  VdVFcn = self.V_sss))

        #Add each pair of expansion chambers
        nCmax = scroll_geo.nC_Max(self.geo)
        # Must have at least one pair
        assert (nCmax>=1)
        
        # Innermost pair at nCmax - 1 starts at inlet state
        
        for alpha in range(nCmax,0,-1): 
            keye1 = 'e1.'+str(alpha)
            keye2 = 'e2.'+str(alpha)
            if alpha == nCmax:
                #It is the outermost pair of compression chambers
                initState = CPState(inletState.Fluid,
                                        dict(T=inletState.T,
                                             D=inletState.rho)
                                        )
            else:
                #It is not the first CV, more involved analysis
                #Assume isentropic compression from the inlet state at the end of the suction process
                T1 = inletState.T
                s1 = inletState.s
                rho1 = inletState.rho
                k = inletState.cp/inletState.cv
                V1 = self.V_sss(2*pi-self.theta_d-1e-14)[0]
                V2 = self.V_e1(0, alpha)[0]*2
                # Mass is constant, so rho1*V1 = rho2*V2
                rho2 = rho1 * V1 / V2
                # Now don't know temperature or pressure, but you can assume
                # it is isentropic to find the temperature
                T2 = optimize.newton(lambda T: PropsSI('S','T',T,'D',rho2,inletState.Fluid)-s1*1000, T1)
                initState = CPState(inletState.Fluid,dict(T=T2,D=rho2)).copy()
            
            # Expansion chambers do not change definition at discharge angle
            disc_becomes_e1 = keye1
            disc_becomes_e2 = keye2
            
            if alpha > 1:
                exists = False # Starts out not being in existence, comes into 
                               # existence at the discharge angle
            else:
                exists = True 
                               
            if alpha == nCmax:
                # It is the innermost pair of chambers, becomes a discharge
                # chamber at the end of the rotation
                becomes_e1 = 'e1.'+str(alpha-1)
                becomes_e2 = 'e2.'+str(alpha-1)
            else:
                # It is not the innermost pair of chambers, becomes another 
                # set of compression chambers at the end of the rotation
                becomes_e1 = 'd1'
                becomes_e2 = 'd2'
                               
            self.add_CV(ControlVolume(key = keye1,
                                      initialState = initState.copy(),
                                      VdVFcn = self.V_e1,
                                      VdVFcn_kwargs = {'alpha':alpha},
                                      discharge_becomes = disc_becomes_e1,
                                      becomes = becomes_e1,
                                      exists = exists))
            
            self.add_CV(ControlVolume(key = keye2,
                                      initialState = initState.copy(),
                                      VdVFcn = self.V_e2,
                                      VdVFcn_kwargs = {'alpha':alpha},
                                      discharge_becomes = disc_becomes_e2,
                                      becomes = becomes_e2,
                                      exists = exists))
            
    def lump_energy_balance_callback(self):
        """
        The callback where the energy balance is carried out on the lumps
        
        Notes
        -----
        Derivation for electrical power of motor:
        
        .. math ::
            
            \\eta _{motor} = \\frac{\\dot W_{shaft}}{\\dot W_{shaft} + \\dot W_{motor}}
            
        .. math ::
            
            {\\eta _{motor}}\\left( \\dot W_{shaft} + \\dot W_{motor} \\right) = \\dot W_{shaft}
            
        .. math::
        
            \\dot W_{motor} = \\frac{\\dot W_{shaft}}{\\eta _{motor}} - \\dot W_{shaft}
        """
        
        #For the single lump
        # HT terms are positive if heat transfer is TO the lump
        Qnet = 0.0
        Qnet -= sum([Tube.Q for Tube in self.Tubes])
        
        self.Qamb = self.ambient_heat_transfer(self.Tlumps[0])
        self.mech.Wdot_losses = 0.9*self.omega/1000.0
        
        # Heat transfer with the ambient; Qamb is positive if heat is being removed, thus flip the sign
        Qnet -= self.Qamb
        
        Qnet += self.mech.Wdot_losses
        # Heat transfer with the gas in the working chambers.  mean_Q is positive
        # if heat is transfered to the gas in the working chamber, so flip the 
        # sign for the lump
        Qnet -= self.HTProcessed.mean_Q
        
        self.Wdot_mechanical = self.Wdot_pv + self.mech.Wdot_losses
        
        #The actual torque required to do the compression [N-m]
        self.tau_mechanical = self.Wdot_mechanical / self.omega * 1000
        
        # 2 Options for the motor losses:
        # a) Constant efficiency
        # b) Based on the torque-speed-efficiency motor
        
#         if self.motor.type == 'const_eta_motor':
#             self.eta_motor = self.motor.eta_motor
#         elif self.motor.type == 'motor_map':
#             # Use the motor map to calculate the slip rotational speed [rad/s]
#             # and the motor efficiency as a function of the torque [N-m]
#             eta, omega = self.motor.apply_map(self.tau_mechanical)
#             self.eta_motor = eta
#             self.omega = omega
#         else:
#             raise AttributeError('motor.type must be one of "const_eta_motor" or "motor_map"')
        
#         #Motor losses [kW]
#         self.motor.losses = 0#self.Wdot_mechanical*(1/self.eta_motor-1)
        
        #Electrical Power
        self.Wdot_electrical = self.Wdot_mechanical# + self.motor.losses
        
        if hasattr(self,'Wdot_i'):
            #Overall isentropic efficiency
            self.eta_oi = self.Wdot_electrical/self.Wdot_i
        
        print('At this iteration')
        print('    Electrical power:', self.Wdot_electrical,'kW')
        print('    Mass flow rate:', self.mdot,'kg/s')
        if hasattr(self,'Wdot_i'):
            print('    Over. isentropic:', self.eta_oi,'-')
        if hasattr(self,'eta_v'):
            print('    Volumetric:', self.eta_v,'-')
        
        #Want to return a list
        return [Qnet]
    
    def S_to_SS(self, FlowPath, **kwargs):
        if self.__before_discharge1__:
            FlowPath.A = 0.0
        else:
            FlowPath.A = scroll_geo.Area_d_dd(2*pi-self.beta, self.geo)
        try:
            mdot = flow_models.IsentropicNozzle(FlowPath.A,
                                               FlowPath.State_up,
                                               FlowPath.State_down)
            return mdot
        except ZeroDivisionError:
            return 0.0
        
    def DA_D(self, FlowPath, X_d=1.0,**kwargs):
        FlowPath.A=X_d*scroll_geo.Area_s_sa(2*pi-self.beta, self.geo)
        try:
            return flow_models.IsentropicNozzle(FlowPath.A,
                                                FlowPath.State_up,
                                                FlowPath.State_down)
        except ZeroDivisionError:
            return 0.0
        
    def step_callback(self,t,h,Itheta):
        """
        Here we test whether the control volumes need to be
        a) Split
        b) Adjusted because you are at the discharge angle
        
        """ 
        #This gets called at every step, or partial step
        self.beta=t
        
        def angle_difference(angle1,angle2):
            # Due to the periodicity of angles, you need to handle the case where the
            # angles wrap around - suppose theta_d is 6.28 and you are at an angles of 0.1 rad
            #, the difference should be around 0.1, not -6.27
            # 
            # This brilliant method is from http://blog.lexique-du-net.com/index.php?post/Calculate-the-real-difference-between-two-angles-keeping-the-sign
            # and the comment of user tk
            return (angle1-angle2+pi)%(2*pi)-pi
        
        def IsAtSplit():
            if t > 3.5 and t < 2*pi-self.theta_d:
                return True
            else:
                return False
            
        disable=False
        
        if t<2*pi-self.theta_d<t+h and self.__before_discharge2__==False:
            #Take a step almost up to the discharge angle using the old definitions
            disable=True
            h=2*pi-self.theta_d-t-1e-10
            self.__before_discharge2__=True
        elif self.__before_discharge2__==True:
            #At the discharge angle
            print('At the discharge angle')
            ########################
            #Reassign chambers
            ########################
            #Find chambers with a discharge_becomes flag
            for key in self.CVs.exists_keys:
                if self.CVs[key].discharge_becomes in self.CVs.keys:
                    #Set the state of the "new" chamber to be the old chamber
                    oldCV=self.CVs[key]
                    #print 'setting',key,'to',oldCV.discharge_becomes
                    if oldCV.exists==True:
                        newCV=self.CVs[oldCV.discharge_becomes]
                        newCV.State.update({'T':oldCV.State.T,'D':oldCV.State.rho})
                        oldCV.exists = False
                        newCV.exists = True
                    else:
                        raise AttributeError("old CV doesn't exist")
            
            self.__before_discharge2__ = False
            self.__before_discharge1__ = True
            
            self.update_existence()
            
            #Re-calculate the CV volumes
            V,dV = self.CVs.volumes(t)
            #Update the matrices using the new CV definitions
            self.T[self.CVs.exists_indices,Itheta]=self.CVs.T
            self.p[self.CVs.exists_indices,Itheta]=self.CVs.p
            self.m[self.CVs.exists_indices,Itheta]=arraym(self.CVs.rho)*V
            self.rho[self.CVs.exists_indices,Itheta]=arraym(self.CVs.rho)
            
            # Adaptive makes steps of h/4 3h/8 12h/13 and h/2 and h
            # Make sure step does not hit any *right* at theta_d
            # That is why it is 2.2e-10 rather than 2.0e-10
            h=2.2e-10
            disable=True
       
        elif self.CVs['sss'].exists and IsAtSplit():
            
            #Build the volume vector using the old set of control volumes (pre-merge)
            V,dV=self.CVs.volumes(t)
            
            print('splitting')
            
            if self.__hasLiquid__==False:

                T = self.CVs['sss'].State.T
                rho = self.CVs['sss'].State.rho
                self.CVs['s1'].State.update({'T':T,'D':rho})
                self.CVs['s2'].State.update({'T':T,'D':rho})
                self.CVs['ss'].State.update({'T':T,'D':rho})
                
                self.CVs['sss'].exists = False
                self.CVs['s1'].exists = True
                self.CVs['s2'].exists = True
                self.CVs['ss'].exists = True
                
                self.update_existence()
                
                #Re-calculate the CV
                V,dV=self.CVs.volumes(t)
                self.T[self.CVs.exists_indices,Itheta] = self.CVs.T
                self.p[self.CVs.exists_indices,Itheta] = self.CVs.p
                self.m[self.CVs.exists_indices,Itheta] = arraym(self.CVs.rho)*V
                self.rho[self.CVs.exists_indices,Itheta] = arraym(self.CVs.rho)
                
            else:
                raise NotImplementedError('no flooding yet')
            
            h = 1e-8
            disable=True 
              
        elif t > 2*pi - self.theta_d:
            self.__before_discharge1__ = False
            disable = False
            
        return disable, h
    
    def pre_run(self):
        """
        Intercepts the call to pre_run and does some scroll processing, then 
        calls the base class function
        """
        
        #Call the base class function        
        PDSimCore.pre_run(self)
        
def Expander():
    
    Ref = 'Nitrogen'
    Tinlet = 300
    pinlet = 2000
    poutlet = 500
    
    SE = ScrollExpander()
    SE.set_scroll_geo(100e-6,2.3,0.005,0.005)
    SE.set_disc_geo('2Arc', r2 = 0)
    
    SE.geo.delta_suction_offset = 0.0e-3
    SE.geo.phi_ie_offset = 0.0
    
    SE.omega = 60*2*pi
    SE.Tamb = 298.0
    
    #Temporarily set the bearing dimensions
    SE.mech = core.struct()
    SE.mech.D_upper_bearing = 0.02540
    SE.mech.L_upper_bearing = 0.031
    SE.mech.c_upper_bearing = 10e-6 #??
    SE.mech.D_crank_bearing = 0.0254
    SE.mech.L_crank_bearing = 0.023
    SE.mech.c_crank_bearing = 10e-6 #??
    SE.mech.D_lower_bearing = 0.01895
    SE.mech.L_lower_bearing = 0.023
    SE.mech.c_lower_bearing = 10e-6 #??
    SE.mech.thrust_ID = 0.05
    SE.mech.thrust_friction_coefficient = 0.02 #Tuned for point A1
    SE.mech.L_ratio_bearings = 5
    SE.mech.mu_oil = 0.008
    SE.mech.scroll_plate_thickness = 0.007
    SE.mech.scroll_density = 8100
    SE.mech.scroll_added_mass = 0
    SE.mech.scroll_plate_diameter = 0.092 #[m]
    m, zcm = SE.calculate_scroll_mass()
    SE.mech.orbiting_scroll_mass = m
    SE.mech.scroll_zcm__thrust_surface = zcm
    
    SE.mech.detailed_analysis = False
    
    SE.h_shell = 10
    SE.A_shell = 0.05
    SE.HTC = 0.5
    
    SE.generator = Motor()
    SE.generator.set_eta(1.0)
    SE.generator.suction_fraction = 0.0
    
    inletState = CPState(Ref,{'T':Tinlet,'P':pinlet})
    T2s = SE.guess_outlet_temp(inletState,poutlet)
    outletState = CPState(Ref,{'T':T2s,'P':poutlet})
    
    SE.geo.delta_radial = 0e-6
    SE.geo.delta_flank = 0e-6

    mdot_guess = inletState.rho*SE.Vdisp/SE.Vratio*SE.omega/(2*pi)
    
    SE.add_tube(Tube(key1 = 'inlet.1',
                     key2 = 'inlet.2',
                     L = 0.03,
                     ID = 0.02,
                     mdot = mdot_guess, 
                     State1 = inletState.copy(),
                     fixed = 1,
                     TubeFcn = SE.TubeCode))
    SE.add_tube(Tube(key1 = 'outlet.1',
                     key2 = 'outlet.2',
                     L = 0.03,
                     ID = 0.02,
                     mdot = mdot_guess, 
                     State2 = outletState.copy(),
                     fixed = 2,
                     TubeFcn = SE.TubeCode))
    
    SE.auto_add_CVs(inletState, outletState)
#    SE.auto_add_leakage(flankFunc = SE.FlankLeakage, 
#                        radialFunc = SE.RadialLeakage)

    FP = FlowPath(key1='outlet.1', 
                  key2='da',
                  MdotFcn=IsentropicNozzleWrapper(),
                  )
    FP.A = pi*0.02**2/4
    SE.add_flow(FP)
    
    SE.add_flow(FlowPath(key1 = 'da', 
                         key2 = 'd1',
                         MdotFcn = SE.DA_D,
                         MdotFcn_kwargs = dict(X_d = 0.5)
                         )
                )
    SE.add_flow(FlowPath(key1 = 'da',
                         key2 = 'd2',
                         MdotFcn = SE.DA_D,
                         MdotFcn_kwargs = dict(X_d = 0.5)
                         )
                )
    
    FP = FlowPath(key1='inlet.2', 
                  key2='ss', 
                  MdotFcn=IsentropicNozzleWrapper(),
                  )
    FP.A = pi*0.007**2/4
    SE.add_flow(FP)
    
    FP = FlowPath(key1='inlet.2', 
                  key2='sss', 
                  MdotFcn=IsentropicNozzleWrapper(),
                  )
    FP.A = pi*0.007**2/4
    SE.add_flow(FP)
    
    SE.add_flow(FlowPath(key1='s1',
                         key2='ss',
                         MdotFcn=SE.S_to_SS))
    SE.add_flow(FlowPath(key1='s2',
                         key2='ss',
                         MdotFcn=SE.S_to_SS))
    
    SE.connect_callbacks(step_callback = SE.step_callback,
                         lumps_energy_balance_callback = SE.lump_energy_balance_callback,
                         endcycle_callback= SE.endcycle_callback)
    
#    beta = np.linspace(0,2*pi,1000)
#    V_e11 = np.array([SE.V_e1(b,1)[0] for b in beta])
#    V_e21 = np.array([SE.V_e2(b,1)[0] for b in beta])
#    V_e12 = np.array([SE.V_e1(b,2)[0] for b in beta])
#    V_e22 = np.array([SE.V_e2(b,2)[0] for b in beta])
#    V_d1 = np.array([SE.V_d1(b)[0] for b in beta])
#    V_d2 = np.array([SE.V_d2(b)[0] for b in beta])
#    V_s1 = np.array([SE.V_s1(b)[0] for b in beta])
#    
#    plt.plot(beta,V_e11)
#    plt.plot(beta,V_e12)
#    plt.plot(beta,V_d1)
#    plt.plot(beta,V_s1)
#    plt.gca().axvline(2*pi-SE.theta_d)
#    plt.show()
    #try:
    SE.eps_cycle = 0.003
    SE.precond_solve(key_inlet='inlet.1',
                     key_outlet='outlet.2',
                     solver_method='RK45',
                     UseNR = False, #If True, use Newton-Raphson ND solver to determine the initial state
                     OneCycle = False,
                     plot_every_cycle= False,
                     hmin = 1e-8
                     )
    #except ValueError as E:
    #    print E
    #    pass
    
if __name__=='__main__':
    Expander()



from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp cimport bool

cpdef bool is_in_vector(string a, vector[string] v)
cpdef double get_map_sd(map[string, double] m, string s)



from PDSim.flow.flow_models import IsentropicNozzleWrapper
from PDSim.flow.flow import FlowPath
from PDSim.core.core import PDSimCore, struct
from PDSim.core.containers import Tube, ControlVolume
from PDSim.core.motor import Motor




if __name__=='__main__':
    from distutils.core import setup
    from distutils.extension import Extension
    from Cython.Distutils import build_ext

    import sys
    sys.argv += ['build_ext','--inplace']

    ext =	Extension("pyclipper", 
                    sources=["pyclipper.pyx", "clipper.cpp"],
                    language="c++",              # this causes Pyrex/Cython to create C++ source
                    include_dirs=["./../include"],
                    )

    setup(
            ext_modules=[ext],
            cmdclass = {'build_ext': build_ext},
    )





from PDSim.misc.datatypes import *
from PDSim.misc.datatypes cimport *

########## Heat transfer callbacks ##########
cdef class HeatTransferCallback(object):
    cdef object core 
    cpdef arraym call(self, double t)

cdef class WrappedHeatTransferCallback(HeatTransferCallback):
    cdef object func
    cpdef arraym call(self, double t)
    
########## Lumps energy balance callbacks ##########
cdef class LumpsEnergyBalanceCallback(object):
    cdef object core 
    cpdef arraym call(self)

cdef class WrappedLumpsEnergyBalanceCallback(LumpsEnergyBalanceCallback):
    cdef object func
    cpdef arraym call(self)
    
########## Step callbacks ##########
cdef class StepCallback(object):
    cdef readonly object disable_adaptive
    cdef readonly double h
    cdef object core
    cpdef double call(self, double t, double h, int i) except *
    
cdef class WrappedStepCallback(StepCallback):
    cdef object func
    cpdef double call(self, double t, double h, int i) except *
    
########## Callbacks container ##########    
cdef class CallbackContainer(object):
    cdef public StepCallback step_callback
    cdef public HeatTransferCallback heat_transfer_callback
    cdef public LumpsEnergyBalanceCallback lumps_energy_balance_callback
    cdef public object endcycle_callback


from __future__ import print_function, absolute_import

from math import pi,sqrt,log
import numpy as np
from ._bearings import calculate_epsilon_short, calculate_epsilon_long

class struct: pass

def thrust_bearing(**kwargs):
    """
    Analysis for the journal bearing
    
    We only take keyword arguments to avoid problems with order of parameters
    Parameters must be specified by name!
    
    Parameters
    ----------
    mu : float
        Friction coefficient [-]
    V : float
        Contact velocity [m/s]        
    N : float
        Thrust force [N]
    
    Notes
    -----
    To derive the speed of contact of the orbiting scroll
    
    .. math::
    
        x = {r_o}\cos \left( {{\phi _{ie}} - \\frac{\pi }{2} - \\theta } \\right)

    .. math::
    
        y = {r_o}\sin \left( {{\phi _{ie}} - \\frac{\pi }{2} - \\theta } \\right)
    
    .. math::
    
        \\frac{{dx}}{{d\\theta }} = {r_o}\sin \left( {{\phi _{ie}} - \\frac{\pi }{2} - \\theta } \\right)\omega 
    
    .. math::
    
        \\frac{{dy}}{{d\\theta }} =  - {r_o}\cos \left( {{\phi _{ie}} - \\frac{\pi }{2} - \\theta } \\right)\omega

    .. math::
    
        \left| v \\right| = \sqrt {{{\left( {\\frac{{dx}}{{d\\theta }}} \\right)}^2} + {{\left( {\\frac{{dy}}{{d\\theta }}} \\right)}^2}}  = {r_o}\omega
        
    But it is quite possible that you do not have hydro-dynamic lubrication, and 
    as a result you can get asperity-asperity contact and much higher friction
    coefficients - see for example Kobayashi et al. "Experimental Study on Journal Bearing
    Characteristics in Reciprocating Compressors for HFC-134a" Purdue Compressor
    Conferences 1998.  http://docs.lib.purdue.edu/cgi/viewcontent.cgi?article=2257&context=icec
    
    Can be off by a factor of as much as 10 times at low
    Sommerfeld number
     
    """
    #Friction coefficient
    mu = kwargs.pop('mu', None)
    V = kwargs.pop('V', None)
    N = kwargs.pop('N', None)
    
    Wdot_loss_thrust = mu * V * N
    return dict(mu = mu,
                V = V,
                Wdot_loss = Wdot_loss_thrust
                )

def journal_bearing(**kwargs):
    
    """
    The necessary calculations for a journal bearing
    
    We only take keyword arguments to avoid problems with order of parameters
    
    Parameters must be specified by name!
    
    Parameters
    ----------
    r_b : float
        Radius of journal [m]
    L : float
        Length of journal [m]
    omega : float
        Rotational speed [rad/s]
    eta_0 : float, or 1d numpy array
        Viscosity of lubricant [Pa-s]
    W : float
        Applied load [N]
    c : float
        Bearing clearance [m]
    
    Returns
    -------
    output_dict : dictionary
        Dictionary of output terms, 
        
    Notes
    -----
    
    Short-bearing analysis is used.  Here we are implementing the equations from 
    the Ph.D. dissertation of Jay Kim (Purdue, 2005).  Knowing :math:`w_r` and 
    the dimensions we can then obtain :math:`\epsilon` using a 1D secant solver
    with the initial guess of :math:`\epsilon=0.5`.  Then we obtain 
    :math:`F_{shear}` for the given value of :math:`\epsilon`. 
    
    .. math::
    
        W_r = \\frac{w_r}{\\eta_0\\omega r_b L}\\left(\\frac{c}{L}\\right)^2 = \\frac{\\epsilon}{4(1-\\epsilon^2)^2}[16\\epsilon^2+\\pi^2(1-\\epsilon^2)]^{1/2}

    .. math::
    
        \mathbf{F}_{shear} = \\frac{F_{shear}}{\\eta_0\\omega r_b L}\\left(\\frac{c}{r_b}\\right) = \\frac{2\\pi}{(1-\\epsilon^2)^{1/2}}\\left[\\left(\\frac{b}{r_b}\\right)^2\\frac{\\epsilon^2}{16(1-\\epsilon^2)}+1\\right]
        
    .. math::
    
        \\mu = \\frac{F_{shear}}{w_r}

    """

    L = kwargs['L']
    r_b = kwargs['r_b']
    W = kwargs['W']
    eta_0 = kwargs['eta_0']
    omega = kwargs['omega']
    c = kwargs['c']
       
    # Short bearing analysis
    Wr_short = W/(eta_0*omega*r_b*L)*(c/L)**2
    
    # Long bearing analysis
    Wr_long = W/(eta_0*omega*r_b*L)*(c/r_b)**2
    
    #Select the first guess value based on the input to the function
    if Wr_long < 5:
        epsilon0_long = 0.01
    else:
        epsilon0_long = 0.8
        
    if Wr_short < 0.8:
        epsilon0_short = 0.01
    else:
        epsilon0_short = 0.8
    
    # Sommerfeld number
    S = eta_0*omega*r_b*L/(pi*W)*(r_b/c)**2
    
    if L/(2*r_b) < 0.5: #Short bearing analysis
        epsilon = calculate_epsilon_short(np.log(Wr_short), epsilon0_short)
        Fshear = 2*pi/np.sqrt(1-epsilon**2)*((L/r_b)**2*epsilon**2/(16*(1-epsilon**2)) +1)*eta_0*omega*r_b*L*(r_b/c)
    
    elif L/(2*r_b) > 2.0: #Infinitely-long analysis
        epsilon = calculate_epsilon_long(np.log(Wr_long), epsilon0_long)
        Fshear = pi/np.sqrt(1-epsilon**2)*(5*epsilon**2+4)/(epsilon**2+2)*eta_0*omega*r_b*L*(r_b/c)
        
    else: #Weighted average of the two to calculate the shear force
        epsilon = calculate_epsilon_long(np.log(Wr_long),epsilon0_long)
        if not (epsilon >= 0 and epsilon <= 1):
            #print 'Wr_long,epsilon,W',Wr_long,epsilon0_long,W
            if epsilon < 0:
                epsilon = 0
            elif epsilon > 1:
                raise ValueError('epsilon [{epsilon:g}] is greater than 1 for W of [{W:g}] N.  Is your journal bearing too small?'.format(epsilon = epsilon, W = W))
        Fshear_long = pi/np.sqrt(1-epsilon**2)*(5*epsilon**2+4)/(epsilon**2+2)*eta_0*omega*r_b*L*(r_b/c)
        epsilon = calculate_epsilon_short(np.log(Wr_short),epsilon0_short)
        assert epsilon >= 0 and epsilon <= 1
        Fshear_short = 2*pi/np.sqrt(1-epsilon**2)*((L/r_b)**2*epsilon**2/(16*(1-epsilon**2)) +1)*eta_0*omega*r_b*L*(r_b/c)
        Fshear = (Fshear_long-Fshear_short)/1.5*(L/(2*r_b)-0.5) + Fshear_short
    
    hmin_over_c = 1-epsilon
    
    h_min = hmin_over_c*c
    
    f = Fshear/W
    
    #Frictional losses [W]
    Wdot_loss = omega*r_b*f*W
    
    return dict(
                S = S,
                f = f,
                c = c,
                Load = W,
                r_b = r_b,
                h_min = h_min,
                epsilon = epsilon,
                omega = omega,
                eta_0 = eta_0,
                Wdot_loss = Wdot_loss
                )
         
def journal_bearing_design(**kwargs):
    
    """
    The necessary calculations for a journal bearing
    
    We only take keyword arguments to avoid problems with order of parameters
    
    Parameters must be specified by name!
    
    Parameters
    ----------
    r_b : float
        Radius of journal [m]
    L : float
        Length of journal [m]
    omega : float
        Rotational speed [rad/s]
    eta_0 : float
        Viscosity of lubricant [Pa-s]
    W : float
        Applied load [N]
    c : float
        Bearing clearance [m]
    design : string or float
        Either one of ``'friction'`` (design for minimum friction) 
        ``'load'`` (design for maximum load) or a floating point 
        value in the range [0,1] that weights the minimum friction
        and maximum load parameters and 0 gives the friction solution.
        Hamrock recommends a value of 0.5 for general applications
        
    Notes
    -----
    One of ``design`` or ``c`` must be provided
    
    Based on the method presented by
    
        Raimondi, A. A., and Boyd, J. (1958) : A Solution for the Finite 
        Journal Bearing and Its Application to Analysis and Design-I, -II, 
        and -III. ASLE Trans., vol. 1, no. I, I- pp. 159-174; II- pp. 175-193; 
        III- pp. 194-209.
    
    And further analysis presented in Hamrock:
    
        In Fig. 11.2 a recommended operating eccentricity ratio, or minimum film
        thickness, is indicated as well as a preferred operating area. The left boundary of
        the shaded zone defines the optimum eccentricity ratio for a minimum coefficient
        of friction, and the right boundary the optimum eccentricity ratio for maximum
        load. The recommended operating eccentricity for general application is midway
        between these two boundaries.

    """

    L = kwargs['L']
    r_b = kwargs['r_b']
    W = kwargs['W']
    eta_0 = kwargs['eta_0']
    omega = kwargs['omega']
    c = kwargs.get('c',None)
    design = kwargs.get('design',None)
        
    L_over_D = L/(2*r_b)
    
    #First based on L/D and the design type, select the hm_over_c
    if design == 'friction':
        hm_over_c = 0.032*L_over_D**2 + 0.32*L_over_D - 0.052
    elif design == 'load':
        hm_over_c = 0.188272*log(L_over_D) + 0.541167
    elif isinstance(design,float):
        hm_over_c_friction = 0.032*L_over_D**2 + 0.32*L_over_D - 0.052
        hm_over_c_load = 0.188272*log(L_over_D) + 0.541167
        hm_over_c = design*hm_over_c_load+(1-design)*hm_over_c_friction
    else:
        raise ValueError('invalid value for parameter "design"')
    
    #Interpolate in the data from to obtain S and rb/c*f
    _hm_c = [0.9,0.8,0.6,0.4,0.2,0.1,0.03]
    _L_D = [1,0.5,0.25]
    _S = [[1.33,0.631,0.264,0.121,0.0446,0.0188,0.00474],[4.31,2.03,0.779,0.319,0.0923,0.0313,0.00609],[16.2,7.57,2.83,1.07,0.261,0.0736,0.0101]]
    _rb_c_f = [[26.4, 12.8, 5.79, 3.22, 1.7, 1.05, 0.514],[85.6, 40.9, 17, 8.1, 3.26, 1.6, 0.61], [322, 153, 61.1, 26.7, 8.8, 3.5, 0.922]]

    from scipy.interpolate import interp2d
    f = interp2d(x=_hm_c,
                 y=_L_D,
                 z=_S,
                 kind = 'linear')
    S = f(hm_over_c,L_over_D)[0]
    f = interp2d(x=_hm_c,
                 y=_L_D,
                 z=_rb_c_f,
                 kind = 'linear')
    rb_c_f = f(hm_over_c,L_over_D)[0]
    
    #r_b/c from bearing number
    rb_c = sqrt((S*pi*W)/(eta_0*omega*r_b*L))
    
    #Friction factor after all that [-]
    f = rb_c_f/rb_c 
    
    h_min = hm_over_c*(r_b/rb_c)
    
    c = r_b/rb_c
    
    Wdot_loss = omega*r_b*f*W
    
    return dict(
                S = S,
                f = f,
                c = c,
                Load = W,
                r_b = r_b,
                h_min = h_min,
                epsilon = 1-hm_over_c,
                omega = omega,
                eta_0 = eta_0,
                Wdot_loss = Wdot_loss
                )
            
if __name__ == '__main__':
    d = journal_bearing(r_b = 0.043/2.0,
                        L = 0.057,
                        c = 5e-5,
                        W = 12950,
                        eta_0 = 0.0086,
                        omega = 301.9402420413402)
    
    print(d['epsilon'])
    
    print('The following is the output from example 11.1 in Hamrock')
    print(journal_bearing_design(r_b = 0.02, 
                        L = 0.04, 
                        design = 'load', 
                        W = 2200, 
                        eta_0 = 0.17, 
                        omega = 3600/60.0*2*pi
                        ))
    



from __future__ import print_function

import wx
import wx.aui
import matplotlib as mpl
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as Canvas
from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as Toolbar
import numpy as np
from CoolProp.CoolProp import Props
import h5py

import matplotlib.pyplot as plt
from cycler import cycler

class Plot(wx.Panel):
    def __init__(self, parent, id = -1, dpi = None, **kwargs):
        wx.Panel.__init__(self, parent, id=id, **kwargs)
        self.figure = mpl.figure.Figure(dpi=dpi, figsize=(2,2))
        self.canvas = Canvas(self, -1, self.figure)
        self.toolbar = Toolbar(self.canvas)
        self.toolbar.Realize()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.canvas,1,wx.EXPAND)
        sizer.Add(self.toolbar, 0 , wx.LEFT | wx.EXPAND)
        self.SetSizer(sizer)

class PlotNotebook(wx.Panel):
    def __init__(self,  Simulation, parent, id = -1, plot_names=None, family = None):
        wx.Panel.__init__(self, parent, id=id)
        self.nb = wx.aui.AuiNotebook(self)
        sizer = wx.BoxSizer()
        sizer.Add(self.nb, 1, wx.EXPAND)
        self.SetSizer(sizer)
        self.Sim=Simulation
        self.family = family
        
        self.build_main_page()        
        
        if plot_names is not None:
            for plot_name in plot_names:
                matched = False
                for name, func in self.plot_buttons:
                    if name == plot_name:
                        func()
                        matched = True 
                #Not matched, quit
                if not matched:
                    raise ValueError("Could not match the name: '"+plot_name+'"')

    def update(self,Simulation):
        """
        Pass in a new Simulation instance
        
        Will clear all plots
        """
        #Update the structure
        self.Sim = Simulation
        while(self.nb.GetPageCount()>1):
            self.nb.DeletePage(1)
        
    def add(self,name="plot"):
        page = Plot(self.nb)
        self.nb.AddPage(page,name)
        page.figure.gca().set_prop_cycle(
            cycler('color', ['r', 'g', 'b', 'y', 'm', 'c']) *
            cycler('linestyle', ['-', '--', '-.'])
            )

        return page.figure
    
    def build_main_page(self):
        page = wx.Panel(self.nb,-1)
        sizer = wx.FlexGridSizer(cols=2)
        label1=wx.StaticText(page,label='Click on the buttons below to add plot')
        sizer.Add(label1)
        self.plot_buttons=[('Stepsize',self.stepsize_theta),
                           ('Volume v. crank angle',self.V_theta),
                           ('Derivative of Volume v. crank angle',self.dV_dtheta),
                           ('Temperature v. crank angle',self.T_theta),
                           ('Pressure v. crank angle',self.p_theta),
                           ('Pressure v. volume',self.p_V),
                           ('Density v. crank angle',self.rho_theta),
                           ('Mass v. crank angle',self.m_theta),
                           ('Mass flow v. crank angle',self.mdot_theta),
                           ('Temperature-pressure',self.temperature_pressure),             
                           ('Heat transfer v. crank angle', self.heat_transfer),
                           ('Initial temperature history', self.initial_temperature_history),
                           ('Lump residuals v. lump temps', self.lumps_residual_v_lump_temps),
                           ('Discharge residual history', self.discharge_residual_history),
                           ]
        self.recip_plot_buttons = [('Valve lift v. crank angle',self.valve_theta)]
        self.scroll_plot_buttons = [('Pressure profile',self.pressure_profile),
                                    ('Axial force v. crank angle',self.axial_force),
                                    ('X-direction force v. crank angle',self.x_direction_force),
                                    ('Y-direction force v. crank angle',self.y_direction_force),
                                    ('Crank pin force magnitude v. crank angle',self.magnitude_force),
                                    ('Gas Torque v. crank angle',self.torque),
                                    ('Force trace', self.force_trace),
                                    ('Force component trace',self.force_component_trace),
                                    ('Radial force', self.radial_force),
                                    ('Tangential force', self.tangential_force)
                                    ]
        for value, callbackfcn in self.plot_buttons:
            btn = wx.Button(page, label = value)
            sizer.Add(btn)
            btn.Bind(wx.EVT_BUTTON, callbackfcn)
        
        if self.family == 'Scroll Compressor':
            more_plot_buttons = self.scroll_plot_buttons
        elif self.family == 'Recip Compressor':
            more_plot_buttons = self.recip_plot_buttons
        else:
            more_plot_buttons = None
        
        if more_plot_buttons is not None:           
            for value, callbackfcn in more_plot_buttons:
                btn = wx.Button(page, label = value)
                sizer.Add(btn)
                btn.Bind(wx.EVT_BUTTON, callbackfcn)
        else:
            print('could not add more buttons particular to current family:', self.family)
            
        page.SetSizer(sizer)
        self.nb.AddPage(page,"Main")
    
    def get_keys(self):
        if isinstance(self.Sim,h5py.File):
            NCV = self.Sim.get('/CVs/N').value
            keys = [self.Sim.get('/CVs/keys/'+str(i)).value for i in range(NCV)]
        else:
            keys = self.Sim.CVs.keys
        return keys

    def get(self, key):
        if isinstance(self.Sim, h5py.File):
            out = self.Sim.get('/' + key).value
        else:
            out=getattr(self.Sim, key)
        return out
        
    def stepsize_theta(self,event=None):
        #Stepsize
        axes = self.add('Stepsize').gca()
        theta = self.get('t')
        h = theta[1::]-theta[0:len(theta)-1]
        axes.semilogy(theta[0:len(theta)-1], h)
        axes.set_ylabel('Stepsize [rad]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
    def V_theta(self, event=None):
        # Volume
        axes = self.add('Volume').gca()
        theta, V = self.get('t'), self.get('V')
        V[V<1e-15]=np.nan
        axes.plot(theta, V.T, lw = 1.5)
        axes.set_ylabel('Volume [cm$^{3}$]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
    def dV_dtheta(self,event=None):
        #Derivative of Volume
        axes = self.add('Vol. Derivative').gca()
        theta, dV = self.get('t'), self.get('dV')
        dV *= 1e6
        dV[np.abs(dV)<1e-15]=np.nan
        axes.plot(theta, dV.T, lw = 1.5)
        axes.set_ylabel('Volume Derivative [cm$^{3}$/rad]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
    def T_theta(self,event=None):
        #Temperature
        axes = self.add('Temperature').gca()
        theta, T = self.get('t'), self.get('T')
        T[T<0.1]=np.nan
        axes.plot(theta, T.T, lw = 1.5)
        axes.set_ylabel('Temperature [K]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def p_theta(self,event=None):    
        #pressure
        axes = self.add('Pressure').gca()
        theta, p = self.get('t'), self.get('p')
        p[p<0.1] = np.nan
        axes.plot(theta, p.T, lw = 1.5)
        axes.set_ylabel('Pressure [kPa]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def p_V(self, event=None):
        #pressure-volume
        axes = self.add('P-V').gca()
        V, p = self.get('V'), self.get('p')
        V *= 1e6
        p[p<0.1] = np.nan
        V[V<1e-15] = np.nan
        axes.plot(V.T, p.T, lw = 1.5)
        axes.set_ylabel('Pressure [kPa]')
        axes.set_xlabel(r'Volume [cm$^{3}$]')
        
    def rho_theta(self,event=None):    
        #density
        axes = self.add('Density').gca()
        theta, rho = self.get('t'), self.get('rho')
        rho[rho<0.1]=np.nan
        axes.plot(theta, rho.T, lw = 1.5)
        axes.set_ylabel('Density [kg/m$^{3}$]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def m_theta(self,event=None):
        #Mass
        axes = self.add('Mass').gca()
        theta, rho, V = self.get('t'), self.get('rho'), self.get('V')
        m = rho*V
        m[m<1e-20]=np.nan
        axes.plot(theta, m.T, lw = 1.5)
        axes.set_ylabel('Mass [kg]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def mdot_theta(self,event=None):    
        #Mass Flow
        axes = self.add('Mdot').gca()
        axes.plot(self.Sim.FlowsProcessed.t,self.Sim.FlowsProcessed.summed_mdot['outlet.1'], lw = 1.5)
        axes.plot(self.Sim.FlowsProcessed.t,self.Sim.FlowsProcessed.summed_mdot['inlet.2'], lw = 1.5)
        axes.plot(self.Sim.FlowsProcessed.t,self.Sim.FlowsProcessed.mean_mdot['outlet.1']*np.ones_like(self.Sim.FlowsProcessed.t), lw = 1.5)
        axes.plot(self.Sim.FlowsProcessed.t,self.Sim.FlowsProcessed.mean_mdot['inlet.2']*np.ones_like(self.Sim.FlowsProcessed.t), lw = 1.5)
        axes.set_xlabel(r'$\theta$ [rad]')
        axes.set_ylabel(r'$\dot m$ [kg/s]')

    def initial_temperature_history(self, event = None):
        axes = self.add('Initial Temperature History').gca()
        if isinstance(self.Sim, h5py.File):
            TTT = self.Sim.get('/solvers/initial_states_history').value
        else:
            TTTT = self.Sim.solvers.initial_states_history
        xx = np.array(list(range(TTT.shape[0])))
        yy = TTT
        axes.plot(xx, yy, 'o-')
        axes.set_xlabel('Iteration Number')
        axes.set_ylabel('Temperature [K]')
        
    def lumps_residual_v_lump_temps(self, event = None):
        axes = self.add('Lump Error History').gca()
        if isinstance(self.Sim, h5py.File):
            Tlumps = self.Sim.get('/solvers/lump_eb_history/Tlumps').value
            lump_eb_error = self.Sim.get('/solvers/lump_eb_history/lump_eb_error').value
        else:
            Tlumps = self.solvers.lump_eb_history.Tlumps
            lump_eb_error = self.solvers.lump_eb_history.lump_eb_error
            
        axes.plot(Tlumps, lump_eb_error, 'o-')
        axes.set_xlabel('Lump temperature [K]')
        axes.set_ylabel('Lump energy balance [kW]')
        
    def discharge_residual_history(self, event = None):
        axes = self.add('Discharge Residual History').gca()
        if isinstance(self.Sim, h5py.File):
            hd = self.Sim.get('/solvers/hdisc_history/hd').value
            hd_error = self.Sim.get('/solvers/hdisc_history/hd_error').value
        else:
            hd = self.solvers.hdisc_history.hd
            hd_error = self.solvers.hdisc_history.hd_error
        axes.plot(hd, hd_error, 'o-')
        axes.set_xlabel('Discharge enthalpy [kJ/kg]')
        axes.set_ylabel('Discharge state error [kJ/kg]')
    
    def valve_theta(self, event = None):
        #valve lift
        if hasattr(self.Sim,'__hasValves__') and self.Sim.__hasValves__:
            axes = self.add('Valves').gca()
            axes.plot(self.Sim.t,self.Sim.xValves[0,:], lw = 1.5)
            axes.plot(self.Sim.t,self.Sim.xValves[2,:], lw = 1.5)
            axes.set_xlabel(r'$\theta$ [rad]')
            axes.set_ylabel(r'Valve lift [m]')
        
    def temperature_pressure(self, event = None):
        #Fluid T-p plot
        axes = self.add('T-P phase').gca()
        if isinstance(self.Sim,h5py.File):
            Fluid = self.Sim.get('/Tubes/0/State1/Fluid').value
            T = self.Sim.get('/T').value.T
            p = self.Sim.get('/p').value.T
        else:
            Fluid = self.Sim.CVs.exists_CV[0].State.Fluid
            T = self.Sim.T
            p = self.Sim.p
            
        #Saturation curve
        Tsat = np.linspace(PropsSI(Fluid,'Tmin')+0.1, PropsSI(Fluid,'Tcrit')-1e-6)
        psat = np.array([PropsSI('P', 'T', T_, 'Q', 1.0, Fluid)/1000 for T_ in Tsat])
        axes.plot(Tsat, psat, lw = 1.5)
        axes.plot(T,p,'.')
        axes.set_xlabel('Temperature [K]')
        axes.set_ylabel(r'Pressure [kPa]')
            
    def heat_transfer(self, event = None):
        axes = self.add('Heat transfer').gca()
        theta, Q = self.get('t'), self.get('Q')
        Q[np.abs(Q)<1e-12]=np.nan
        axes.plot(theta, Q.T, lw = 1.5)
        axes.plot(theta,self.Sim.HTProcessed.summed_Q[0:self.Sim.Ntheta],lw=2)
        axes.set_ylabel(r'$\dot Q$ [kW]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
    def axial_force(self, event = None):
        #Axial force
            
        axes = self.add('Axial Force').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Fz = self.Sim.get('/forces/Fz').value.T
            summed_Fz = self.Sim.get('/forces/summed_Fz').value.T
        else:
            theta = self.Sim.t
            Fz = self.Sim.forces.Fz.T
            summed_Fz = self.Sim.forces.summed_Fz.T
        
        Fz[np.abs(Fz)<1e-12]=np.nan
        axes.plot(theta,Fz, lw = 1.5)
        axes.plot(theta,summed_Fz,lw=4)
        axes.set_ylabel(r'$F_z$ (only from the applied gas) [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
    def x_direction_force(self, event = None):
        #x-direction force
            
        axes = self.add('X Force').gca()
        
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Fx = self.Sim.get('/forces/Fx').value.T
        else:
            theta = self.Sim.t
            Fx = self.Sim.forces.Fx.T
            
        Fx[np.abs(Fx)<1e-12]=np.nan
        axes.plot(theta,Fx, lw = 1.5)
        axes.set_ylabel(r'$F_x$ [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def y_direction_force(self, event = None):
        #y-direction force

        axes = self.add('Y Force').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Fy = self.Sim.get('/forces/Fy').value.T
        else:
            theta = self.Sim.t
            Fy = self.Sim.forces.Fy.T
            
        Fy[np.abs(Fy)<1e-12]=np.nan
        axes.plot(theta,Fy, lw = 1.5)
        axes.set_ylabel(r'$F_y$ [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
    def force_component_trace(self, event = None):
        #trace of force components
            
        axes = self.add('Force trace').gca()
        if isinstance(self.Sim,h5py.File):
            Fx = self.Sim.get('/forces/Fx').value.T
            Fy = self.Sim.get('/forces/Fy').value.T
        else:
            Fx = self.Sim.forces.Fx.T
            Fy = self.Sim.forces.Fy.T
        axes.plot(Fx,Fy, lw = 1.5)
        axes.set_ylabel(r'$F_x$ [kN]')
        axes.set_xlabel(r'$F_y$ [kN]')            
        
    def force_trace(self, event = None):
        #trace of force components
            
        axes = self.add('Force trace').gca()

        if isinstance(self.Sim,h5py.File):
            Fx = self.Sim.get('/forces/summed_Fx').value.T
            Fy = self.Sim.get('/forces/summed_Fy').value.T
        else:
            Fx = self.Sim.forces.summed_Fx.T
            Fy = self.Sim.forces.summed_Fy.T
            
        Fx[np.abs(Fx)<1e-12]=np.nan
        Fy[np.abs(Fy)<1e-12]=np.nan
        
        axes.plot(Fx,Fy, lw = 1.5)
        axes.set_ylabel(r'$F_x$ [kN]')
        axes.set_xlabel(r'$F_y$ [kN]')
        
    def magnitude_force(self, event = None):
        #Crank pin force magnitude
            
        axes = self.add('Shaft force magnitude').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Fm = self.Sim.get('/forces/Fm').value.T
        else:
            theta = self.Sim.t
            Fm = self.Sim.forces.Fm.T
            
        Fm[np.abs(Fm)<1e-12] = np.nan
        axes.plot(theta, Fm, lw = 1.5)
        axes.set_ylabel(r'$F_m$ [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
    def radial_force(self, event = None):
        #Radial force magnitude
            
        axes = self.add('Radial force magnitude').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Fr = self.Sim.get('/forces/Fr').value.T
            mean_Fr = self.Sim.get('/forces/mean_Fr').value
            summed_Fr = self.Sim.get('/forces/summed_Fr').value
        else:
            theta = self.Sim.t
            Fr = self.Sim.forces.Fr.T
            mean_Fr = self.Sim.forces.mean_Fr
            summed_Fr = self.Sim.forces.summed_Fr
            
        Fr[np.abs(Fr)<1e-12]=np.nan
        axes.plot(theta,Fr, lw = 1.5)
        axes.plot(theta,mean_Fr*np.ones_like(theta),'k--', lw = 1.5)
        axes.plot(theta,summed_Fr,lw=4)
        axes.set_ylabel(r'$F_r$ (only from the applied gas) [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
        
    def tangential_force(self, event = None):
        #Tangential force magnitude
            
        axes = self.add('Tangential force magnitude').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            Ft = self.Sim.get('/forces/Ft').value.T
            mean_Ft = self.Sim.get('/forces/mean_Ft').value
            summed_Ft = self.Sim.get('/forces/summed_Ft').value
        else:
            theta = self.Sim.t
            Ft = self.Sim.forces.Ft.T
            mean_Ft = self.Sim.forces.mean_Ft
            summed_Ft = self.Sim.forces.summed_Ft
            
        Ft[np.abs(Ft)<1e-12]=np.nan
        axes.plot(theta,Ft, lw = 1.5)
        axes.plot(theta, mean_Ft*np.ones_like(theta), 'k--', lw = 1.5)
        axes.plot(theta,summed_Ft, lw = 4)
        axes.set_ylabel(r'$F_t$ (only from the applied gas) [kN]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(self.get_keys(), loc='upper right', bbox_to_anchor=(1, 1),
          ncol=2, fancybox=True, shadow=True)
    
        
    def torque(self, event = None):
        #Torque
        axes = self.add('Torque').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('/t').value
            tau = self.Sim.get('/forces/tau').value.T
            mean_tau = self.Sim.get('/forces/mean_tau').value
        else:
            theta = self.Sim.t
            tau = self.Sim.forces.tau.T
            mean_tau = self.Sim.forces.mean_tau
            
        tau[np.abs(tau)<1e-12]=np.nan
        axes.plot(theta,tau, lw = 1.5)
        axes.plot(theta, mean_tau*np.ones_like(theta),'k--', lw = 1.5)
        axes.set_ylabel(r'$\tau$ [kN-m]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
    def pressure_profile(self, event = None):
        #pressure profiles
            
        axes = self.add('Pressure profiles').gca()
        if isinstance(self.Sim,h5py.File):
            theta = self.Sim.get('summary/theta_profile').value
            p1 = self.Sim.get('summary/p1_profile').value.T
            p2 = self.Sim.get('summary/p2_profile').value.T
        else:
            theta = self.Sim.t
            p1 = self.Sim.summary.p1_profile
            p2 = self.Sim.summary.p2_profile
            
        axes.plot(theta, p1, lw = 1.5, label='s1-c1.x-d1-ddd')
        axes.plot(theta, p2, lw = 1.5, label='s2-c2.x-d2-ddd')
        axes.set_ylabel(r'Pressure [kPa]')
        axes.set_xlabel(r'$\theta$ [rad]')
        
        xmin,xmax = axes.get_xlim()
        axes.set_xlim(xmin, xmin + (xmax-xmin)*1.5)
        axes.legend(loc='upper right', bbox_to_anchor=(1, 1), ncol=2, 
                    fancybox=True, shadow=True)
         
    
def debug_plots(Comp, plotparent=None, plot_names = None):
    # Build a new frame, not embedded
    app = wx.App(False)
    frame = wx.Frame(None, -1,'Plotter')
    notebook = PlotNotebook(Comp, frame, plot_names=plot_names)
    frame.Show()
    app.MainLoop()

if __name__ == "__main__":
    print("File for doing plots.  Don't run this file")




from CoolProp.State import State as StateClass
from CoolProp.State cimport State as StateClass

from CoolProp.constants_header cimport parameters

from libcpp cimport bool

from PDSim.flow.flow import FlowPathCollection
from PDSim.flow.flow cimport FlowPathCollection

from PDSim.misc.datatypes cimport arraym
from PDSim.misc.datatypes import arraym
    
cdef class TubeCollection(list):
    cdef dict _Nodes
    cdef arraym harray, parray, Tarray
    
    cpdef update_existence(self, int NCV)
    cpdef arraym get_h(self)
    cpdef arraym get_p(self)
    cpdef arraym get_T(self)
    cpdef dict get_Nodes(self)
    cpdef update(self)
    
cdef class Tube(object):
    cdef bytes m_key1,m_key2
    cdef public int fixed
    cdef public StateClass State1, State2
    cdef public object TubeFcn
    cdef public double Q_add,alpha,L,ID,OD,mdot,Q
    cdef public bool exists
    cdef public int i1,i2
    
cdef class ControlVolume(object):
    cdef public long keyIndex
    cdef bytes m_key, m_discharge_becomes
    cdef public object becomes
    cdef public object V_dV
    cdef public dict V_dV_kwargs
    cdef public object ForceFcn
    cdef public bint exists
    cdef public StateClass State

cdef class CVScore(object):
    cdef list array_list
    
    # Other variables
    cdef int state_vars, N
    cdef double omega
    
    # Storage arrays that are always required
    cdef public arraym T,p,h,rho,V,dV,cp,cv,m,v,dpdT_constV,Q
    
    # Property derivative arrays
    cdef public arraym summerdm, summerdT, drhodtheta, dTdtheta, dmdtheta, property_derivs
    
    cpdef update_size(self, int N)
    cdef build_all(self, int N)
    cdef free_all(self)
    cpdef copy(self)
    cpdef calculate_flows(self, FlowPathCollection Flows)
    cpdef just_volumes(self, list CVs, double theta)

cdef class CVArrays(CVScore):
    cpdef properties_and_volumes(self, list CVs, double theta, int state_vars, arraym x)
    cpdef calculate_derivs(self, double omega, bint has_liquid)

cdef class ControlVolumeCollection(object):
    cdef readonly list keys, CVs, indices, exists_keys, exists_indices, exists_CV
    cdef readonly dict Nodes
    cdef readonly int N, Nexist
    cdef public CVArrays arrays

    cpdef add(self, ControlVolume CV)
    cpdef rebuild_exists(self)
    cpdef updateStates(self, str name1, arraym array1, str name2, arraym array2)
    cpdef volumes(self, double theta, bint as_dict = *)
    cpdef at(self, int i)
    cpdef get(self, parameters key, double factor=*)




import numpy as np
cimport numpy as np

import cython
cimport cython

from PDSim.scroll.common_scroll_geo cimport geoVals, HTAnglesClass

from libc.math cimport sqrt,sin,cos,tan,atan2,acos,floor,M_PI as pi,pow,atan

cpdef double fxA(double rb, double phi, double phi0)
cpdef double fyA(double rb, double phi, double phi0)

cpdef double theta_d(geoVals geo) except *

cpdef double radial_leakage_area(double theta, geoVals geo, long key1, long key2, int location = *) except *
cdef radial_leakage_angles(double theta, geoVals geo, long key1, long key2, double *angle_min, double *angle_max)

cpdef int getNc(double theta, geoVals geo) except *

cpdef HTAnglesClass HT_angles(double theta, geoVals geo, key)

cpdef tuple SA(double theta, geoVals geo, bint poly=*, bint use_offset = *, double Vremove = *)
cpdef dict SA_forces(double theta, geoVals geo, bint poly = *, bint use_offset = *)

cpdef tuple S1(double theta, geoVals geo, bint poly = *, double theta_0_volume = *, bint use_offset = *)
cpdef dict S1_forces(double theta, geoVals geo, bint poly = *, double theta_0_volume =*, bint use_offset = *)

cpdef tuple S2(double theta, geoVals geo, bint poly = *, double theta_0_volume = *)
cpdef dict S2_forces(double theta, geoVals geo, bint poly = *, double theta_0_volume =*)

cpdef tuple C1(double theta, int alpha, geoVals geo, bint poly=*)
cpdef dict C1_forces(double theta, int alpha, geoVals geo, bint poly = *)

cpdef tuple C2(double theta, int alpha, geoVals geo, bint poly=*)
cpdef dict C2_forces(double theta, int alpha, geoVals geo, bint poly = *)

cpdef tuple D1(double theta, geoVals geo, bint poly=*)
cpdef dict D1_forces(double theta, geoVals geo, bint poly = *)

cpdef tuple D2(double theta, geoVals geo, bint poly=*)
cpdef dict D2_forces(double theta, geoVals geo, bint poly = *)

cpdef tuple DD(double theta, geoVals geo, bint poly=*)
cpdef dict DD_forces(double theta, geoVals geo, bint poly=*)

cpdef tuple DDD(double theta, geoVals geo, bint poly=*) 
cpdef dict DDD_forces(double theta, geoVals geo, bint poly=*) 

cpdef CVcoords(CVkey, geoVals geo, double theta)

cpdef double phi_s_sa(double theta, geoVals geo)

@cython.locals(iter=cython.int,phi_os=cython.double,phi_o0=cython.double,phi_ie=cython.double,phi_i0=cython.double,change=cython.double,eps=cython.double,f=cython.double,x1=cython.double,x2=cython.double,x3=cython.double,y1=cython.double,y2=cython.double,phi=cython.double,alpha=cython.double)
cpdef double phi_d_dd(double theta, geoVals geo)

cpdef double Area_d_dd(double theta, geoVals geo)
cpdef double Area_s_sa(double theta, geoVals geo)
cpdef double Area_s_s1_offset(double theta, geoVals geo)



from PDSim.flow import flow_models
cimport PDSim.flow.flow_models as flow_models

from PDSim.flow.flow_models import FlowFunction
from PDSim.flow.flow_models cimport FlowFunction

from PDSim.flow.flow import FlowPath
from PDSim.flow.flow cimport FlowPath

from PDSim.scroll import scroll_geo

from PDSim.scroll.common_scroll_geo import geoVals
from PDSim.scroll.common_scroll_geo cimport geoVals
    
cdef class _Scroll(object):
    cdef public geoVals geo
    cdef public double theta
    cdef public double HTC
    
    cpdef dict __cdict__(self)
    cpdef double SA_S(self, FlowPath FP)
    cpdef double Discharge(self,FlowPath FP)
    cpdef double Inlet_sa(self, FlowPath FP)
    cpdef double RadialLeakage(self, FlowPath FP, double t = *)
    cpdef double FlankLeakage(self, FlowPath FP, int Ncv_check = *)
